;
;   Here are the options.
;
(declare (usual-integrations))
;
(define *connect-groups?* #t)
(define *log* #t)
;
(define *last-runtime* 0)
;
(define (this-time)
   (let* ((t (round (* (runtime) 1000))) (r (- t *last-runtime*)))
      (set! *last-runtime* t)
      r))
;
(define (logit a)
   (when *log* (display-stuff  " [" (this-time) "]" 'newline "; " a)))
;
(define *graph* '())
;
(define *seqno* 0)
;
(define *the-asc* '()) ; for debugging; see (the-asc)
;
(define *global-assoc* (list (cons '*IO* '()) (cons 'GLOBAL 0)))
;
(define (add-to-environment foo)
   (let ((v (assoc (car foo) *global-assoc*)))
      (if v
         (set-cdr! v (cdr foo))
         (set-cdr! *global-assoc* (cons foo (cdr *global-assoc*)))))
   #t)
;
;   In CScheme (and real Scheme, by the way) brackets ([]) are not legal
;   syntax for symbols.  So we use colons instead, after much gnashing of
;   teeth.  But internally, we still use brackets.
;
(define *eps-brackets* (string->symbol "[]"))
(define *eps-immediate* (string->symbol "[!]"))
;
(define (new-group-num)
   (let ((oval (graph-maxgroupval *graph*)))
      (graph-maxgroupval-set! *graph* (+ oval 1))
      oval))
;
(define (node-name n)
   (cons-symbol (node-type n) (node-seq n)))
;
(define (new-node-name)
   (set! *seqno* (+ *seqno* 1))
   *seqno*)
;
(define (add-node-to-graph nnode)
   (graph-nodes-set! *graph* (cons nnode (graph-nodes *graph*))))
;
(define (new-node s)
   (let ((nnode (make-node)))
      (node-seq-set! nnode (new-node-name))
      (node-type-set! nnode s)
      (add-node-to-graph nnode)
      nnode))
;
;   We keep track of the orbit that a place belongs to so that we can
;   minimize the length of the state vector.
;
(define *current-orbit* 0)
;
(define *next-orbit* 1)
;
(define (new-place)
   (let ((nnode (new-node 'P)))
      (node-orbit-set! nnode *current-orbit*)
      nnode))
;
(define (new-place-unadded)
   (let ((nnode (make-node)))
      (node-seq-set! nnode (new-node-name))
      (node-type-set! nnode 'P)
      (node-orbit-set! nnode *current-orbit*)
      nnode))
;
(define (new-node-unadded t)
   (let ((nnode (make-node)))
      (node-seq-set! nnode (new-node-name))
      (node-type-set! nnode t)
      nnode))
;
;   Epsilon places don't get orbits.
;
(define (new-eps-place)
   (let ((nnode (new-place)))
      (add-to-label! nnode *eps-brackets*)
      (node-orbit-set! nnode #f)
      nnode))
;
(define (new-func label)
   (let ((nnode (new-node 'F)))
      (node-label-set! nnode (list label))
      nnode))
;
(define (new-trans)
   (new-node 'T))
;
(define (new-reg)
   (new-node 'R))
;
(define (new-reg-from t d)
   (let* ((n (new-reg))
          (e2 (connect-df d n 0)))
      (connect t n)
      (edge-outpos-set! e2 1)
      n))
;
(define (new-join)
   (new-node 'J))
;
(define (new-join-from . args)
   (let ((n (new-join)))
      (for-each (lambda (arg)
              (connect-df arg n 0))
         args)
      n))
;
(define (new-trans-from-place p)
   (let ((nnode (new-trans)))
      (connect p nnode)
      nnode))
;
(define (new-place-from-trans t)
   (let ((nnode (new-place)))
      (connect t nnode)
      nnode))
;
(define (connect n1 n2)
   (let ((e (new-edge)))
      (edge-innode-set! e n1)
      (edge-outnode-set! e n2)
      (node-outedges-set! n1 (cons e (node-outedges n1)))
      (node-inedges-set! n2 (cons e (node-inedges n2)))
      (edge-type-set! e 'P)
      e))
;
(define (connect-df n1 n2 w)
   (let ((e (connect (df-check n1) n2)))
      (edge-type-set! e 'D)
      (edge-outpos-set! e w)
      e))
;
(define (connect-if-head n1 n2)
   (when n1 (connect n1 n2)))
;
(define (remove-if expr lst)
   (let loop ((lst lst) (sofar '()))
      (if lst
         (if (expr (car lst))
	    (loop (cdr lst) sofar)
            (loop (cdr lst) (cons (car lst) sofar)))
         (reverse sofar))))
;
;   Set this flag before you remove a bunch of nodes where you're
;   not planning to run through the structure of the graph for a
;   while.  This should take (remove-extra-epsilon-places) from
;   being quadratic to being linear.
;
(define *caching* #f)
;
(define (start-remove-cache)
   (when (not (memq 'c *debug-flags*))
      (set! *caching* #t)))
;
(define (clean-remove-cache)
   (set! *caching* #f)
   (graph-nodes-set! *graph*
      (remove-if (lambda (a) (not (node-type a))) (graph-nodes *graph*))))
;
(define (remove e l)
   (if l
      (if (eq? e (car l))
         (remove e (cdr l))
         (cons (car l) (remove e (cdr l))))
      '()))
;
(define (remove-edge e)
   (let ((inn (edge-innode e))
         (onn (edge-outnode e)))
      (when *caching*
         (edge-type-set! e #f))
      (node-inedges-set! onn (remove e (node-inedges onn)))
      (node-outedges-set! inn (remove e (node-outedges inn)))))
;
(define (remove-node n)
; (display-stuff "Removing node " (node-name n) 'newline)
   (let ((ine (node-inedges n))
         (one (node-outedges n)))
      (if *caching*
         (node-type-set! n #f)
         (graph-nodes-set! *graph* (remove n (graph-nodes *graph*))))
      (node-inedges-set! n '())
      (node-outedges-set! n '())
      (let loop ((nl (append ine one)))
         (if nl
            (begin
               (remove-edge (car nl))
               (loop (cdr nl)))))))
;
(define (move-edge-head e p1 p2)
   (when p1
      (node-outedges-set! p1 (remove e (node-outedges p1))))
   (if p2
      (begin
         (edge-innode-set! e p2)
         (node-outedges-set! p2 (cons e (node-outedges p2))))))
;
(define (move-edge-tail e p1 p2)
   (when p1
      (node-inedges-set! p1 (remove e (node-inedges p1))))
   (edge-outnode-set! e p2)
   (node-inedges-set! p2 (cons e (node-inedges p2))))
;
(define (new-edge)
   (let ((e (make-edge)))
      (edge-inpos-set! e 0)
      (edge-outpos-set! e 0)
      e))
;
(define (add-to-label! n s)
   (if (null? s)
      (merror "Not a legal label type" s))
   (node-label-set! n (cons s (node-label n))))
;
;   We want to make darn sure that the message displays, so we do it
;   ourselves before calling error.
;
(define (merror a b)
   (display-stuff 'newline " *** ERROR *** " a ": ")
   (display-stuff b 'newline)
   (show-context)
   (error b a))
;
(define (node-type? n p)
   (eq? (node-type n) p))
;
(define (edge-type? e p)
   (eq? (edge-type e) p))
;
(define (df-check w)
   (if (number? w)
      (new-func w)
      (if (and (node? w) (or (node-type? w 'F)
                             (node-type? w 'P)
                             (node-type? w 'R)
                             (node-type? w 'J)))
         w
         (merror "Expected df node" (node-type w)))))
;
(define *global-sub-net* '())
;
(define (build-graph expr)
   (set! *eval-log* '())   
   (let ((car-ga (car *global-assoc*))
         (cdr-ga (cdr *global-assoc*))
         (returned #f))
      (set! *basic-place* '())
      (set! *graph* (make-graph))
      (set! *sgraph* #f)
      (set! *current-orbit* 0)
      (set! *next-orbit* 1)
      (set! *global-sub-net* '())
      (let ((start-place (new-place)))
         (graph-maxgroupval-set! *graph* 0)
         (logit "Parsing the expression")
         (set! returned (build-expr-top expr start-place))
	 (when (and (= (length (graph-nodes *graph*)) 1)
		    (cdr returned)
		    (circuit? (cadr returned)))
	    (build-expr-top (add-wires-to-circuit (cadr returned))
	       start-place))
         (propogate-io-info (graph-nodes *graph*))
         (when *verify-df-wires?*
            (logit "Resolving value wires")
            (connect-value-wires))
         (fire-simplify start-place)
         (fix-simplify-keys)
         (if *connect-groups?*
            (connect-groups))
         (remove-extra-epsilon-places)
         (minimize-orbits))
      (set! *global-assoc* (cons car-ga cdr-ga)))
   (set! *eval-log* '())   
   *graph*)
;
(define (propogate-io-info nl)
   (when nl
      (let ((n (car nl)))
         (when (and (node-type? n 'T) (node-input-output? n))
            (for-each (lambda (e)
                         (let ((nn (edge-innode e)))
                            (when (and (node-type? nn 'P)
				     (not (edge-ignore-timing? e)))
                               (node-input-output?-set! nn
                                  (node-input-output? n)))))
               (node-inedges n)))
         (propogate-io-info (cdr nl)))))
;
(define *eval-log* '())
(define *eval-keep* 25)
(define *eval-current* 0)
;
;   Turns an arbitrary expression into a string of maximum length
;   supplied.
;
(define (check-length s n)
   (if (> (string-length s) n)
      (if (< n 4)
         "..."
         (string-append (substring s 0 (- n 3)) "..."))
      s))
;
(define (expr->string expr n)
   (check-length (expr->string* expr n) n))
;
(define (expr->string* expr n)
   (cond
      ((< n 0) "...")
      ((number? expr) (number->string expr))
      ((string? expr) expr)
      ((symbol? expr) (symbol->string expr))
      ((eq? expr #t) "#t")
      ((eq? expr #f) "()")
      ((pair? expr)
         (let loop ((expr expr) (s "(") (n (- n 1)))
	    (if (< n 0) s
               (let ((cars (expr->string* (car expr) n)))
                  (if (pair? (cdr expr))
		     (loop (cdr expr) (string-append s cars " ")
			   (- n (string-length cars) 1))
		     (if (null? (cdr expr))
		        (string-append s cars ")")
			(string-append s cars " . "
			   (expr->string* (cdr expr) (- n (string-length s) 4))
			   ")")))))))
      ((vector? expr) (string-append "#"
		         (expr->string* (vector->list expr) (- n 1))))
      (else "???")))
;
(define (show-context)
   (when *eval-log*
      (display-stuff 'newline "   Recently evaluated:" 'newline)
      (let loop ((n 8) (ex *eval-log*))
         (when (and (> n 0) ex)
            (display-stuff (expr->string (caar ex) 78) 'newline)
            (loop (- n 1) (cdr ex))))))
;
(define (build-expr-top expr start-place)
   (set! *eval-log* '())
   (set! *eval-current* 0)
   (start-remove-cache)
   (let ((r (build-expr-primitives 'complete (list '() expr)
	       start-place *global-assoc*)))
      (clean-remove-cache)
      r))
;
(define (build-expr expr place asc)
   (DBG 'e (display-stuff "; " (expr->string expr 76) 'newline))
   (set! *eval-log* (cons (cons expr asc) *eval-log*))
   (set! *eval-current* (+ 1 *eval-current*))
   (when (> *eval-current* (* 2 *eval-keep*))
      (set! *eval-log* (my-sublist *eval-log* 0 *eval-keep*))
      (set! *eval-current* *eval-keep*))
   (if (and expr (pair? expr))
      (build-expr-cons (car expr) (cdr expr) place asc)
      (build-expr-atoms expr place asc)))
;
(define (fix-colon-to-brackets s)
   (let* ((s (symbol->string s))
          (sl (string-length s)))
      (cons-symbol "[" (substring s 1 (- sl 1)) "]")))
;
;   We want to use floating point everywhere.  If we encounter a float,
;   we want to `fix' it.
;
(define (fix-the-flonum n)
   (ascii->number (reverse (string->list (number->string n))) 1 0))
;
(define (build-expr-atoms expr place asc)
   (if (null? expr)
      (list place)
      (if (number? expr)
	 (if (flo:flonum? expr)
	    (list place (fix-the-flonum expr))
            (list place expr))
         (if (symbol? expr)
            (let ((val (assq expr asc)))
               (if val
                  (list place (cdr val))
;
;   If the first and last character is :, we translate them to [].
;   If the last character is + or -, we treat it as an annotate.
;   If the last character is [], we treat it as a delay.
;
                  (let* ((s (symbol->string expr))
                         (sl (string-length s))
                         (lastchar (string-ref s (- sl 1))))
                     (case lastchar
                        (#\+ 
                           (build-expr-primitives 'annotate
						  (list expr) place asc))
                        (#\-
                           (build-expr-primitives 'annotate
						  (list expr) place asc))
                        (#\]
                           (build-expr-primitives 'delay
						  (list expr) place asc))
                        (#\:
                           (build-expr
			     (list 'delay
			        (fix-colon-to-brackets expr)) place asc))
                        (else
                           (merror "That symbol wasn't defined" expr))))))
            (merror "Got something we didn't expect" expr)))))
;
(define (minimal-net pair)
   (let ((n (car pair)))
      (and (eq? n (cadr pair))
        (null? (node-inedges n))
        (null? (node-outedges n))
        (null? (node-label n)))))
;
(define (make-key size)
   (when (or (> size 1) (< size 0))
      (merror "Only key sizes of 0 or 1 currently supported" size))
   (let ((p (new-eps-place)))
      (let ((k
               (cons 'key (cons p (if (> size 0) (list (new-join)) '())))))
         (graph-keys-set! *graph* (cons k (graph-keys *graph*)))
         k)))
;
;   Registers are very similar.
;
(define (make-reg)
   (let ((p (new-join)))
      p))
;
(define *basic-place* '())
;
(define (simple-expr exp asc)
   (if (integer? exp)
      exp
      (if (number? exp)
         (if (flo:flonum? exp)
	    (fix-the-flonum exp)
	    exp)
         (let ((v (and (symbol? exp) (assq exp asc))))
            (if v (cdr v)
               (let* ((place (or *basic-place*
			        (begin
			           (set! *basic-place* (new-place-unadded))
				   *basic-place*)))
                      (r (build-expr exp place asc)))
                  (if (eq? (car r) place)
                     (cadr r)
                     (merror "! expected simple expression here" exp))))))))
;
(define *orb-eval-asc* '())
;
(define (orb-eval exp)
   (simple-expr exp *orb-eval-asc*))
;
(define (my-local-eval expr)
   (eval expr (the-environment)))
;
(define (verify-key key asc)
   (let ((val (simple-expr key asc)))
      (if (and val (pair? val) (eq? (car val) 'key)) val
         (merror "Not a key!" key))))
;
(define (verify-wire w asc)
   (let ((val (simple-expr w asc)))
      (if (and val (pair? val) (eq? (car val) 'wire)) val
         (merror "Not a wire!" w))))
;
(define (bang-key place t key args asc)
   (when (not (= (length args) (length (cddr key))))
      (merror "Wrong number of arguments given for bang of key" args))
   (let* ((res (if args (build-expr (car args) place asc) (list place)))
          (kp (cadr key))
          (p (car res)))
      (if (not (eq? p place))
         (merror "Too complex an expr inside bang-place" 0))
      (connect t kp)
      (when (> (length args) 0)
         (let ((r (new-reg-from t (cadr res))))
            (connect-df r (caddr key) 0)))
      (list place)))
;
(define (verify-register key asc)
   (let ((val (simple-expr key asc)))
      (if (and (node? val) (node-type? val 'J))
         val
         (merror "Not a register!" key))))
;
(define (my-set! place args asc)
   (let* ((j (verify-register (car args) asc))
          (nv (build-expr (cadr args) place asc))
          (t (new-trans-from-place (car nv)))
          (p2 (new-place-from-trans t))
          (r (new-reg-from t (cadr nv))))
      (connect-df r j 0)
      (list p2)))
;
(define (atomic-read place args asc)
   (let ((j (simple-expr (car args) asc)))
      (let* ((t (new-trans-from-place place))
             (p2 (new-place-from-trans t))
             (r (new-reg-from t j))
             (nj (new-join-from r)))
         (list p2 nj))))
;
(define (bang-key-value place key wire asc)
   (let* ((key (verify-key key asc))
          (kp (cadr key)))
      (when (not (= 1 (length (cddr key))))
         (merror "Wrong number of arguments given for bang of key" args))
      (add-to-label! place wire)
      (connect place kp) ; we'll fix this later
      (connect-df place (caddr key) 0)
      (list place)))
;
;   Here we handle a query inside a setplace.
;
(define (setval-query place key)
   (let* ((kp (cadr key))
          (kj (caddr key)))
      (connect-df kj place 0)
      (connect kp place)
      (list place)))
;
(define (query-key place t key)
   (let* ((kp (cadr key))
          (res
             (if (cddr key)
                (let ((r (new-reg-from t (caddr key))))
                   (list r))
                '())))
      (connect kp t)
      (cons place res)))
;
(define (merge-patch-assoc p1 p2 asc)
   (for-each (lambda (p)
                (when (eq? (safe-cadr p) p1)
                   (set-car! (cdr p) p2)))
      asc))
;
;   Only special-case 0.  We also spawn off any additional orbit
;   numbers as necessary.  We populate them as full as we can.
;
;   We also cheat a little; if one of the exprs is very simple (a
;   number constant or it is on the assoc list), then we don't even
;   allocate it a place.
;
;   Note that this function is called a *lot*, so little optimizations
;   here can improve parsing.
;
(define (build-expr-par args place asc)
   (let ((oorbit *next-orbit*) (oc *current-orbit*))
      (set! *next-orbit* (* (length args) *next-orbit*))
      (let loop ((co *current-orbit*) (prelim '())
                 (nontrivs '()) (todo args))
         (if todo
            (let ((exp (car todo)))
               (set! *current-orbit* co)
               (if (integer? exp)
                  (loop co (cons (list exp) prelim) nontrivs (cdr todo))
                  (let ((v (and (symbol? exp) (assq exp asc))))
                     (if v
                        (loop co (cons (list (cdr v)) prelim) nontrivs
                           (cdr todo))
                        (let* ((p (new-place-unadded))
                               (t (cons p (build-expr (car todo) p asc))))
                           (if (minimal-net t)
                              (loop co (cons (cddr t) prelim)
                                 nontrivs (cdr todo))
                              (loop (+ co oorbit) (cons (cddr t) prelim)
                                 (cons t nontrivs) (cdr todo))))))))
            (let ((numargs (length args))
                  (numdontreturn
                     (length (remove-if (lambda (a) (cadr a)) nontrivs))))
               (set! *current-orbit* oc)
               (set! *next-orbit* oorbit)
               (unless (or (= numargs numdontreturn)
                           (= numdontreturn 0))
                  (display-stuff args 'newline)
                  (merror "In parallel exe, either all or none must return"
                     args))
               (case (length nontrivs)
                  (0 (cons place (reverse prelim)))
                  (1 (let ((oneplace (caar nontrivs)))
                        (if (node-label oneplace)
                           (let ((t1 (new-trans-from-place place)))
                              (add-node-to-graph oneplace)
                              (connect t1 oneplace))
                           (begin
                              (merge-patch-assoc oneplace place asc)
                              (merge-nodes place oneplace)))
                        (cons (cadar nontrivs) (reverse prelim))))
                  (else (let ((t1 (new-trans-from-place place))
                              (t2 (if (< numdontreturn numargs)
                                     (new-trans) '())))
                           (for-each (lambda (e)
                                       (add-node-to-graph (car e))
                                       (connect t1 (car e))
                                       (if t2
                                          (connect-if-head (cadr e) t2)))
                             (reverse nontrivs))
                           (cons (if t2 (new-place-from-trans t2) '())
                              (reverse prelim))))))))))
;
(define (build-expr-cons proc args place asc)
   (unless (symbol? proc)
      (merror "We only handle symbols in the first position." proc))
   (build-expr-primitives proc args place asc))
;
(define (<< a b)
   (if (< b 1) a
      (<< (* a 2) (- b 1))))
;
(define (>> a b)
   (if (< b 1) a
      (>> (/ a 2) (- b 1))))
;
(define (primitive-version proc)
   (if (number? proc) (lambda () proc)
      (case proc
         (+ +) (- -) (/ quotient) (* *) (<< <<) (>> >>)
         (= (lambda (a b) (if (= a b) 1 0)))
         (> (lambda (a b) (if (> a b) 1 0)))
         (< (lambda (a b) (if (< a b) 1 0)))
         (>= (lambda (a b) (if (>= a b) 1 0)))
         (<= (lambda (a b) (if (<= a b) 1 0)))
         (<> (lambda (a b) (if (= a b) 0 1)))
         (odd (lambda (a) (if (odd? a) 1 0)))
         (even (lambda (a) (if (even? a) 1 0)))
         (not (lambda (a) (if (= a 0) 1 0)))
         (and (lambda (a b) (if (or (= a 0) (= b 0)) 0 1)))
         (or (lambda (a b) (if (and (= a 0) (= b 0)) 0 1)))
         (? (lambda () 'unknown))
         (else '()))))
;
(define (known? a)
   (and (pair? (cdr a)) (number? (cadr a))))
;
(define (all-known? a)
   (and (pair? (cdr a))
      (let loop ((nl (cdr a)))
         (or (null? nl) (and (number? (caar nl)) (loop (cdr nl)))))))
;
(define (all-known proc pars place)
   (if (and (eq? place (car pars))
          (null? (remove-if (lambda (a) (and (pair? a) (number? (car a))))
                    (cdr pars))))
      (let ((p (primitive-version proc)))
	 (if p
            (list place (apply (primitive-version proc) (map car (cdr pars))))
           '()))
      '()))
;
;   Since `=' is used so often, we speed it up.
;
(define (all-known= pars place)
   (if (and (eq? place (car pars))
          (= (length pars) 3))
      (let ((n1 (cadr pars)) (n2 (caddr pars)))
         (if (and (pair? n1) (pair? n2) (number? (car n1)) (number? (car n2)))
            (list place (if (= (car n1) (car n2)) 1 0))
	    '()))
      '()))
;
(define (bigger-vector v ns)
   (let* ((n (+ 10 (* ns 3)))
	  (nv (make-vector n '())))
      (let loop ((n (- (vector-length v) 1)))
         (when (>= n 0)
	    (vector-set! nv n (vector-ref v n))
	    (loop (- n 1))))
      nv))
;
(define (check-v-grow sm)
   (let* ((ns (abs (caaar (sm-newstates sm)))))
      (when (>= ns (vector-length (sm-state-v sm)))
         (sm-state-v-set! sm (bigger-vector (sm-state-v sm) ns))
         (sm-nstate-v-set! sm (bigger-vector (sm-nstate-v sm) ns)))))
;
(define (sm-assoc s v)
   (let ((n (abs (car s))))
      (and (< n (vector-length v))
         (assoc s (vector-ref v n)))))
;
(define (sm-push val s v)
   (let ((n (abs (car s))))
      (vector-set! v n (cons val (vector-ref v n)))))
;
(define (build-expr-sm name binds exprs pars place asc)
   (let ((sm (make-sm))
         (oorbit *current-orbit*)
         (nextorbit *next-orbit*))
      (sm-vars-set! sm (map car binds))
      (sm-state-v-set! sm (make-vector 10 '()))
      (sm-nstate-v-set! sm (make-vector 10 '()))
      (sm-newstates-set! sm
         (list (list (map car (cdr pars)) place
                     (cons (cons name sm) asc) *next-orbit*)))
      (check-v-grow sm)
      (sm-push (car (sm-newstates sm))
         (caar (sm-newstates sm)) (sm-nstate-v sm))
      (sm-body-set! sm (cons 'begin exprs))
      (let loop ((exits '()))
         (if (sm-newstates sm)
            (let* ((newstate (car (sm-newstates sm)))
                   (ignore (sm-newstates-set! sm (cdr (sm-newstates sm))))
                   (state (car newstate))
                   (place (cadr newstate))
                   (asc (caddr newstate))
                   (orb (nth 3 newstate))
                   (donestate (sm-assoc state (sm-state-v sm)))
                   (res (if donestate
                           (begin
                              (try-merge-nodes (cdr donestate) place)
                              (list '()))
                           (begin
                              (set! *current-orbit* (node-orbit place))
                              (set! *next-orbit* orb)
			      (let loop ((asc asc))
			         (if (memq (caar asc) (sm-vars sm))
				    (loop (cdr asc))
                                    (build-expr (sm-body sm) place
                                       (append (map cons (sm-vars sm) state)
                                          asc))))))))
               (when (not donestate)
		  (sm-push (cons state place) state (sm-state-v sm)))
               (loop (if (car res) (cons res exits) exits)))
            (begin
               (set! *current-orbit* oorbit)
               (set! *next-orbit* nextorbit)
               (or-join exits))))))
;
(define (call-sm sm args place asc)
   (let* ((state (map car (cdr args)))
          (asc
	     (let loop ((asc asc))
	        (if (memq (caar asc) (sm-vars sm))
		   (loop (cdr asc))
   	           (append (map (lambda (v a)
                                  (if (number? a)
                                     (cons v a)
                                     (merror "Unknown val in sm" a)))
                            (sm-vars sm) state) asc))))
          (donestate (sm-assoc state (sm-state-v sm))))
      (if donestate
         (try-merge-nodes (cdr donestate) place)
	 (begin
            (sm-newstates-set! sm (cons (list state (virgin-place place) asc
                                           *next-orbit*)
                                     (sm-newstates sm)))
            (check-v-grow sm)
            (sm-push (car (sm-newstates sm)) state (sm-nstate-v sm))))
      (list '())))
;
(define (or-join res)
   (let ((rs (remove-if (lambda (a) (null? (car a))) res)))
      (if rs
         (if (= (length rs) 1)
            (car rs)
            (let* ((p1 (new-place))
                   (j1 (if (cdar rs) (new-join) '())))
               (unless (apply = (map length rs))
                  (merror "All returns must yield same number of values" rs))
               (for-each (lambda (a)
                       (if j1
                          (let ((t1 (new-trans-from-place (car a))))
                             (connect-df (new-reg-from t1 (cadr a)) j1 0)
                             (connect t1 p1))
                          (try-merge-nodes p1 (car a))))
                  rs)
               (cons p1 (if j1 (list j1) '()))))
         (list '()))))
;
(define (find-level t asc)
   (let loop ((asc asc))
      (if asc
         (let ((v (caar asc)))
            (if (if (symbol? v)
                    (eq? 'GLOBAL v)
                    (and (pair? v) (eq? (cdr v) t) (eq? (car v) 'level)))
                  (cdar asc)
                  (loop (cdr asc))))
	 (merror "Oops; no global???" t))))
;
;   We also handle annotations of the form (!) and (?).
;
(define (do-annotate p1 p2 t labs asc)
   (cons p2 (remove-if null? (map (lambda (a)
                                     (do-annot a p1 t asc)) labs))))
;
;   This is where we handle burst mode.
;
;   FIXME FIXME FIXME for (complete (...)) . . . need to do each signal
;   separately, rather than all together as we do here.
;
(define (handle-burst burst p asc)
   (let ((np (new-place)))
      (recurse-burst burst p asc np '())
      np))
;
(define (recurse-burst burst p asc np basc)
   (let ((jp (if (> (length burst) 1) #f np)))
      (let loop ((basc basc) (bt burst))
         (if bt
            (let* ((a (car bt))
                   (nburst (remove-if (lambda (b) (eq? a b)) burst))
                   (seen? (or (assoc nburst basc)))
                   (cp (or (safe-cdr seen?) jp))
                   (tp (try-annot a p asc cp))
                   (nbasc (recurse-burst nburst tp asc np basc)))
               (loop (if seen? nbasc (cons (cons nburst tp) nbasc))
                  (cdr bt)))
            basc))))
;
(define (try-annot a p asc jp)
   (let ((io-annot #f)) ; yuck; someone save me from such ugly code!
      (when (symbol? a)
         (set! a (lookup-rise-fall-symbol a asc))
         (when (pair? a)
            (set! io-annot (cdr a))
            (set! a (car a))))
      (let loop ((el (node-outedges p)))
         (if el
            (let ((t (edge-outnode (car el))))
               (if (and (node-type? t 'T)
                          (assoc a (node-label t)))
                  (let ((np (find-output-node-of-type 'P t)))
                     (if (and jp (not (eq? jp np)))
                        (merror "Oops; subset problem with burst" 0)
                        np))
                  (loop (cdr el))))
            (let ((t (new-trans-from-place p)))
               (label-trans t a (find-level a asc))
               (node-input-output?-set! t io-annot)
               (if jp
                  (begin
                     (connect t jp)
                     jp)
                  (new-place-from-trans t)))))))
;
(define (label-trans t a lev)
   (let ((as (assoc a (graph-t-assoc *graph*))))
      (if as
         (set-cdr! as (cons (cons t lev) (cdr as)))
         (graph-t-assoc-set! *graph*
            (cons (list a (cons t lev)) (graph-t-assoc *graph*))))
      (add-to-label! t (cons a lev))))
;
(define (trans-exists? a)
   (assoc a (graph-t-assoc *graph*)))
;
(define (g-annotate p1 p2 t labs asc)
   (if (symbol? labs)
      (g-annotate p1 p2 t (list 'annotate labs) asc)
      (if (and labs (eq? (car labs) 'annotate))
         (do-annotate p1 p2 t (cdr labs) asc)
         (do-annotate p1 p2 t (list labs) asc))))
;
(define (lookup-rise-fall-symbol a asc)
   (let* ((s (symbol->string a))
          (sl (string-length s))
          (fc (string-ref s (- sl 1))))
      (if (or (eq? fc #\+) (eq? fc #\-))
         (let ((w-pair (assq (string->symbol (substring s 0 (- sl 1))) asc)))
            (if (and w-pair (pair? (cdr w-pair)) (eq? (cadr w-pair) 'wire))
;  (begin (display-stuff "Looked up " a " and found " w-pair " so returning "
;               (cons-symbol (caddr w-pair) fc) 'newline)
               (cons
                  (cons-symbol (caddr w-pair) fc)
                  (check-io-in-environment (cdr w-pair) asc))
               (merror "No such wire" a)))
         a)))
;
(define (do-annot a p t asc)
;  (display-stuff "Doing an annotate on " a 'newline)
   (let ((io-annot #f)) ; yuck; someone save me from such ugly code!
      (when (symbol? a)
         (set! a (lookup-rise-fall-symbol a asc))
         (when (pair? a)
            (set! io-annot (cdr a))
            (set! a (car a))))
      (if (or (symbol? a) (not (assoc (cadr a) asc)))
         (begin
            (node-input-output?-set! t io-annot)
            (label-trans t a (find-level a asc))
            '())
         (let* ((k (verify-key (cadr a) asc))
                (r (if (eq? (car a) '!)
                      (bang-key p t k (cddr a) asc)
                      (if (eq? (car a) '?)
                         (query-key p t k)
                         (merror "Bad symbol in first pos of annotation"
                            (car a))))))
            (if (eq? (car r) p)
               (if (pair? (cdr r)) (cadr r) '())
               (merror "cannot create or destroy transitions in ! or ?" 0))))))
;
(define (unpair a)
   (if (pair? a) (car a) a))
;
(define (add-group vars nn asc)
   (if vars
      (let ((v (car vars))
            (r (cdr vars)))
         (if (symbol? v)
            (let ((val (assq v asc)))
               (if val
                  (if (pair? (cdr val))
                     (case (cadr val)
                        (wire (add-group (cons (cons-symbol v '+)
                                            (cons (cons-symbol v '-) r))
                                 nn asc))
                        (key (add-group (cons (list '! v)
                                           (cons (list '? v) r)) nn asc))
                        (else (merror "Hmm, strange value in addgroup" 0)))
                     (merror "Hmm, strange value in addgroup" 0))
               (add-group r nn (cons (cons (cons 'level
                                              (unpair
					         (lookup-rise-fall-symbol
						    v asc)))
                                        nn) asc))))
            (if (pair? v)
;
;   This needs fixing to handle complete keys.  (Is this the bug we found
;   Friday?)
;
;              (add-group r nn (cons (cons (cons 'level v) nn) asc))
               (merror "Wrong type for complete" v))))
      asc))
;
(define (build-expr-primitives proc args place asc)
   (case proc
      (=
         (let* ((pars (build-expr-par args place asc)))
            (or (all-known= pars place)
               (let ((f (new-func proc)))
                  (map (lambda (res pos)
                          (connect-df (car res) f pos))
                     (cdr pars) (iota-l pars))
                  (list (car pars) f)))))
      ((+ - / * >> << < > <= >= <> not and or even odd read write)
         (let* ((pars (build-expr-par args place asc)))
            (or (all-known proc pars place)
               (let ((f (new-func proc)))
                  (map (lambda (res pos)
                          (connect-df (car res) f pos))
                     (cdr pars) (iota-l pars))
                  (list (car pars) f)))))
      (begin
         (if args
            (let loop ((args args) (place place))
               (if place
                  (if args
                     (if (cdr args)
                        (loop (cdr args)
                              (car (build-expr (car args) place asc)))
                        (build-expr (car args) place asc))
                     place)
                  (merror "Intermediate expr doesn't terminate" args)))
            (list place)))
      (if
         (let ((res1 (build-expr (car args) place asc)))
            (if (known? res1)
               (if (= (cadr res1) 0)
                  (build-expr (caddr args) place asc)
                  (build-expr (cadr args) place asc))
               (let* ((t1 (new-trans-from-place (car res1)))
                      (t2 (new-trans-from-place (car res1)))
                      (p1 (new-place-from-trans t1))
                      (p2 (new-place-from-trans t2))
                      (r2 (build-expr (cadr args) p1 asc))
                      (r3 (build-expr (caddr args) p2 asc)))
                  (connect-df (cadr res1) t1 1)
                  (edge-inpos-set! (connect-df (cadr res1) t2 1) 1)
                  (or-join (list r2 r3))))))
      (guard
         (let* ((t
                 (build-expr-par (remove-if null? (map car args)) place asc))
                (booleans
                   (let loop ((t (cdr t)) (args (map car args)))
                      (if args
                         (if (car args)
                            (cons (car t) (loop (cdr t) (cdr args)))
                            (cons '(1) (loop t (cdr args))))
                         '())))
                (p (car t))
                (tails
                   (map (lambda (bool trans expr)
                           (if (and (number? (car bool)) (= (car bool) 0))
                              (list '())
                              (if (and (number? (car bool)) (= (car bool) 1))
                                 (let ((r (build-expr-primitives 'begin
					     (cons trans expr)
                                             (temp-violate p) asc)))
                                    (unviolate p)
                                    r)
                                 (let* ((t1 (new-trans-from-place p))
                                        (p1 (new-place-from-trans t1)))
                                    (when (not (number? (car bool)))
                                       (connect-df (car bool) t1 1))
                                    (if (and trans expr)
                                       (build-expr-primitives 'begin expr
                                          (car (g-annotate p p1 t1 trans asc))
                                          asc)
                                       (begin
                                          (if trans
                                             (g-annotate p p1 t1 trans asc))
                                          (build-expr-primitives 'begin expr
                                             p1 asc)))))))
                      booleans (map safe-cadr-lambda args)
		               (map safe-cddr args))))
                (or-join tails)))
      (case
         (let* ((cv (simple-expr (car args) asc)))
	    (let loop ((argl (if (vector? (cadr args))
			        (vector-ref (cadr args) cv)
				(cdr args)))
		       (tails '()))
	       (if argl
		  (loop (cdr argl)
                     (if (= (caar argl) cv)
			(cons (build-expr-primitives 'begin
			         (cdar argl) place asc)
		           tails)
		        tails))
		  (or-join tails)))))
      (let
         (let ((labeled (symbol? (car args))))
            (when labeled
               (begin
                  (set! asc (cons (list (car args)) asc))
                  (set! labeled (car asc))
                  (set! args (cdr args))))
            (let* ((pars (build-expr-par (map cadr (car args)) place asc))
                   (cmplx ;  do we need this? (or (not (all-known? pars)))
                       (or (and labeled (cdr pars))
                              (not (virgin-place-notime? (car pars)))))
                   (t2 (if cmplx (new-trans-from-place (car pars)) '()))
                   (p2 (if cmplx (new-place-from-trans t2) (car pars)))
                   (reslist (map (lambda (arg)
                                    (if labeled
                                       (new-join-from
                                          (new-reg-from t2 (car arg)))
                                       (car arg)))
                               (cdr pars))))
               (for-each (lambda (res arg)
                            (set! asc (cons (cons (car arg) res) asc)))
                  reslist (car args))
               (when labeled
                  (set-cdr! labeled (cons p2 reslist)))
               (build-expr-primitives 'begin (cdr args) p2 asc))))
      (sm
         (let* ((pars (build-expr-par (map cadr (cadr args)) place asc)))
            (when (not (all-known '+ pars place))
               (merror "Must only use known values in sm" pars))
            (build-expr-sm (car args) (cadr args) (cddr args) pars place asc)))
      (annotate
         (let* ((t1 (new-trans-from-place place))
                (p2 (new-place-from-trans t1)))
           (do-annotate place p2 t1 args asc)))
      (burst
         (list (handle-burst args place asc)))
      (!
         (build-expr-primitives 'annotate (list (cons '! args)) place asc))
      (?
         (build-expr-primitives 'annotate (list (cons '? args)) place asc))
      (forever
         (build-expr (append '(let loop ()) args '((loop))) place asc))
      (delay
         (add-to-label! place (car args))
         (list place))
      (void
         (list '()))
      (par
         (list (car (build-expr-par args place asc))))
      (value
         (add-to-label! place (car args))
         (list place place))
      (!-value
         (bang-key-value place (car args) (cadr args) asc))
      (setvalue
         (let ((res (build-expr (cadr args) place asc)))
            (connect-df (cadr res) place 0)
            (add-to-label! (car res) (car args))
            (list (car res))))
      (?-setvalue
         (let ((k (verify-key (car args) asc)))
            (add-to-label! place (cadr args))
            (connect-df (caddr k) place 0)
            (connect (cadr k) place)
            (list place)))
      (dflabel
         (let ((nr (if (cdr args) (build-expr (cadr args) place asc) '()))
               (j (new-join)))
            (add-to-label! j (car args))
            (if nr
               (begin
                  (connect-df (cadr nr) j 0)
                  (list (car nr) j))
               (list place j))))
      (make-key
         (let ((key-arity (if args (car args) 0)))
            (list place (make-key key-arity))))
      (make-wire
         (list place (make-wire args asc)))
      (make-place
         (list place (make-place args asc)))
      (make-transition
         (list place (make-transition args asc)))
      (make-edges
         (list place (make-edges '() args asc #f)))
      (net
         (make-net place args asc)
	 (list '()))
      (complete
         (let ((nn (new-group-num))
               (vars (car args))
               (expr (cons 'begin (cdr args))))
            (if vars
               (build-expr expr place (add-group vars nn asc))
               (build-expr expr place (cons (cons 'GLOBAL nn) asc)))))
      (the-asc
         (set! *the-asc* asc)
         (list place))
      (process
         (build-expr-primitives 'complete (list '() (cons 'forever args)) place asc))
      (complete-par
         (build-expr-primitives 'complete (list '() (cons 'par args)) place asc))
      (make-reg (list place (make-reg-args args asc)))
      (set! (my-set! place args asc))
      (atomic-read (atomic-read place args asc))
      (when
         (build-expr-primitives 'if
	    (list (car args) (cons 'begin (cdr args)) '())
               place asc))
;      (munge-env
;         (set! *global-assoc* asc)
;         (list place))
      (with-wires
         (build-expr (with-wires args #f) place asc))
      (with-places
         (build-expr (with-places args) place asc))
      (with-transitions
         (build-expr (with-transitions args) place asc))
      (global-wires-def
         (build-expr (with-wires (list args) #t) place asc))
      (wire-init
         (let* ((w (verify-wire (car args) asc))
                (v (nth 2 w)))
            (if (eq? v 'unknown)
               (merror "Initial value of wire not known" (car args))
               (list place v))))
      (wire-init-known?
         (let* ((w (verify-wire (car args) asc))
                (v (nth 2 w)))
            (list place (if (eq? v 'unknown) 0 1))))
      (wire-init-set!
         (if (pair? (car args))
            (build-expr
               (cons 'begin (map (lambda (a) (cons 'wire-init-set! a)) args))
               place asc)
            (let* ((w (verify-wire (car args) asc))
                   (nv (simple-expr (cadr args) asc)))
               (wire-init-set! w nv)
               (list place))))
      (connect
         (let ((b (verify-ckt? (car args) asc))
               (c (verify-ckt? (cadr args) asc))
               (in/out (caddr args))
	       (back-init (if (> (length args) 3) (cadddr args) #f)))
            (when (not (= 1 (macro-inputs b) (macro-outputs b)))
               (merror "Buffer arg to connect must have one input/output"
                  (car args)))
            (list place (connect-buffers b c in/out back-init))))
      (combinational
         (build-expr (build-eval-circ args) place asc))
      (mirror
         (let ((c (verify-ckt? (car args) asc)))
            (list place (mirror-ckt c))))
      (choice
         (build-expr `(case 1 ,@(map (lambda (a) `(1 ,@a)) args)) place asc))
      (die-if
         (build-expr `(case 1 (1) ,(append '(1) args '((void)))) place asc))
      (receptive-wire
         (add-receptiveness args asc))
      (ignore-timing
         (mark-timing-ignored place)
	 (list place))
      (eval-parse
         (set! *orb-eval-asc* asc)
	 (my-local-eval (cons 'begin args))
         (list place))
      (expand
         (set! *orb-eval-asc* asc)
	 (build-expr (my-local-eval (cons 'begin args)) place asc))
      (eval-exec
         (list place))
      (normalize
         (let* ((ckt (verify-ckt? (car args) asc))
		(name (macro-name ckt))
		(margs (macro-formals ckt))
		(ins (my-sublist margs 0 (macro-inputs ckt)))
		(outs (my-sublist margs (macro-inputs ckt) (macro-outputs ckt)))
		(defn (normalize-internal
		         `(with-wires global ,(append ins outs)
			    (the-asc)
			    ,(cons name (append ins outs (cdr args))))
			 ins outs #f)))
	    (list place
               (list 'macro '() margs defn (macro-inputs ckt)
	          (macro-outputs ckt)))))
      (else
         (let ((val (assq proc asc)))
            (if val
               (if (expr-macro? (cdr val))
                  (do-macro (cdr val) args place asc)
                  (if (sm? (cdr val))
                     (call-sm (cdr val) (build-expr-par args place asc)
                        place asc)
                     (let ((pars (build-expr-par args place asc)))
                        (if (and (virgin-place? (car pars))
                                 (null? args))
                           (try-merge-nodes (cadr val) place)
                           (let ((t2 (new-trans-from-place (car pars))))
                              (for-each (lambda (res where pos)
                                         (if (eq? (car res) where) '()
                                           (connect-df (new-reg-from t2
                                                          (car res))
                                              where pos)))
                                 (cdr pars) (cddr val) (iota-l pars))
                              (connect t2 (cadr val))))
                        (list '()))))
               (merror "Proc not handled." proc))))))
;
(define (add-wires-to-circuit ckt)
   (let* ((name (macro-name ckt))
	  (margs (macro-formals ckt))
	  (ins (my-sublist margs 0 (macro-inputs ckt)))
	  (outs (my-sublist margs (macro-inputs ckt) (macro-outputs ckt))))
      `(with-wires global ,(append ins outs)
         ,(cons name (append ins outs)))))
;
(define (mark-timing-ignored p)
   (let* ((t (find-input-node-of-type 'T p)))
      (let loop ((el (node-inedges t)))
         (if el
	    (if (and (node-type? (edge-innode (car el)) 'P)
		   (eq? (node-orbit (edge-innode (car el)))
			(node-orbit p)))
               (edge-ignore-timing?-set! (car el) #t)
	       (loop (cdr el)))
	    (merror "Oops; couldn't find node/edge in ignore-timing" 0)))))
;
(define (make-reg-args args asc)
   (let ((r (make-reg)))
      (when args
         (node-label-set! r (list (simple-expr (car args) asc))))
      r))
;
;   This function takes two circuits---a buffer circuit and `real'
;   (usually combinational) circuit---and builds a new macro that
;   hooks up the buffer circuits to all inputs (in/out true) or
;   outputs of the real circuit.  The result is a macro; the given
;   name is '() (it will be given a name by def.)
;
;   We would also like to maintain the appropriate names, and this
;   would almost work, except that we want to make sure the new wires
;   are unique.  So we cons together a somewhat long name.
;
;   The output should look like:
;
;   (macro () (args of c) body (in-count of c) (out-count of c))
;
;   where body is
;
;   (with-wires (a@ b@ c@) (par
;
;   and then either the circuit followed by the buffers (for output
;   buffering) or the buffers followed by the circuit (for input
;   buffering.)
;
(define (connect-buffers b c in/out back-init)
   (let* ((ink (macro-inputs c)) (outk (macro-outputs c))
          (args (macro-formals c))
          (in (my-sublist args 0 ink)) (out (my-sublist args ink outk))
          (new-wires (map (lambda (a) (cons-symbol a '@)) (if in/out in out)))
          (extra (my-sublist args (+ ink outk) 1000)))
      (list 'macro '() args
         (list 'complete '()
            (list 'with-wires new-wires
               (cons 'begin
	          (if (and back-init (not in/out))
		    (map (lambda (i o)
		            `(when (wire-init-known? ,o)
		               (wire-init-set! ,i (wire-init ,o))))
		       new-wires out)
	          '()))
               (cons 'par
                  (if in/out
                     (append
                        (map (lambda (i o) (list (macro-name b) i o))
                           in new-wires)
                        (list (append (list (macro-name c))
                                 new-wires out extra)))
                     (append
                        (list (append (list (macro-name c))
                                 in new-wires extra))
                        (map (lambda (i o) (list (macro-name b) i o))
                           new-wires out))))))
         ink outk)))
;
;   This function creates the mirror of a circuit simply by swapping inputs
;   and outputs.
;
(define (mirror-ckt c)
   (let* ((ink (macro-inputs c)) (outk (macro-outputs c))
          (args (macro-formals c))
          (in (my-sublist args 0 ink)) (out (my-sublist args ink outk))
          (extra (my-sublist args (+ ink outk) 1000)))
      (list 'macro '() (append out in extra)
         (cons (macro-name c) args)
         outk ink '*mirror*)))
;
;   This function adds receptiveness.  First it makes sure that it has
;   precisely two wires as arguments.  Then, it creates four transitions,
;   one for up and one for down on each wire, and connects them with an
;   immediate place.  It allocates a new `group' number and performs an
;   annotate.  That is all that is needed; the connections will be made
;   by other parts of the system.  We return `void', as it is more like
;   a real parallel buffer than anything else.
;
;   We only add the transitions and places if some other corresponding
;   place/transition already exists.  CHANGED; now we always add, and
;   rely on a post-processing phase to remove them if necessary.
;
(define (add-receptiveness args asc)
   (let ((in-w (verify-wire (car args) asc))
         (out-w (verify-wire (cadr args) asc))
         (gn (new-group-num)))
;      (when (or (eq? (check-io-in-environment in-w asc) '*input*)
;                (eq? (check-io-in-environment out-w asc) '*output*))
;         (merror "Oops; receptive place has input or output problem!" 0))
      (for-each (lambda (a)
                   (let ((wni (cons-symbol (wire-name in-w) a))
                         (wno (cons-symbol (wire-name out-w) a)))
                      (when #t
;                     (when (or (trans-exists? wni) (trans-exists? wno))
                         (let* ((t1 (new-trans))
                                (p1 (new-place-from-trans t1))
                                (t2 (new-trans-from-place p1)))
                            (add-to-label! p1 *eps-immediate*)
                            (node-orbit-set! p1 #f)
                            (node-input-output?-set! t1 '*output*)
                            (node-input-output?-set! t2 '*input*)
                            (label-trans t1 wni gn)
                            (label-trans t2 wno gn)))))
         '(+ -)))
   (list #f))
;
;   Now we `fire' the graph if possible, removing places and tokens
;   as we go, if we can.
;
;   We then scan the entire graph, removing any places that have neither
;   inputs nor outputs.  (These can be generated during fire-simplify.)
;
(define (fire-simplify p)
   (start-remove-cache)
   (fire-simplify-aux p)
   (let loop ((pl (graph-nodes *graph*)))
      (when pl
         (when (and (node-type? (car pl) 'P)
                  (= 0 (length (node-inedges (car pl)))
                       (length (node-outedges (car pl)))))
            (remove-node (car pl)))
         (loop (cdr pl))))
   (clean-remove-cache))
;
(define (fire-simplify-aux p)
   (let ((done #f))
      (if (and (= (length (node-inedges p)) 0)
               (= (length (node-outedges p)) 1)
               (eq? '() (node-label p)))
         (let ((t (edge-outnode (car (node-outedges p)))))
            (if (and (= (length (node-inedges t)) 1)
                     (null? (node-label t)))
               (let ((pl (map edge-outnode-lambda (node-outedges t))))
                  (if (null? (remove-if (lambda (a)
                                           (node-type? a 'P))
                                 pl))
                     (begin
                        (remove-node p)
                        (remove-node t)
                        (let loop ((pl pl))
                           (if pl
                              (begin
                                 (fire-simplify-aux (car pl))
                                 (loop (cdr pl)))
                              (set! done '#t)))))))))
      (if (not done)
         (node-marked-set! p #t))))
;
;   This function makes a new value wire.  Quite a lot of
;   connections.
;
(define (new-vwire name)
   (let* ((p-valid (new-place))
          (t-inv (new-trans-from-place p-valid))
          (p-inv (new-place-from-trans t-inv))
          (t-valid (new-trans-from-place p-inv))
          (do-valid (new-eps-place))
          (die1 (new-trans-from-place do-valid))
          (do-inv (new-eps-place))
          (die2 (new-trans-from-place p-inv))
          (do-read (new-eps-place))
          (ok-read (new-trans-from-place do-read))
          (w (make-vwire)))
      (new-place-from-trans die1) ; these two may not be necessary
      (new-place-from-trans die2) ; but better safe than sorry
      (connect t-valid p-valid)
      (connect do-valid t-valid)
      (connect do-inv t-inv)
      (connect p-valid die1)
      (connect do-read die2)
      (connect ok-read p-valid)
      (connect p-valid ok-read)
      (node-marked-set! p-inv #t)
      (vwire-join-set! w (new-join))
      (vwire-begwrite-set! w do-valid)
      (vwire-endwrite-set! w do-inv)
      (vwire-begread-set! w do-read)
      (vwire-invalidt-set! w t-inv)
      (add-to-label! (vwire-join w) name)
      w))
;
;   This is where we make a wire.
;
(define *wire-seq-no* 1000)
;
(define (next-wire-seq)
   (set! *wire-seq-no* (+ 1 *wire-seq-no*))
   *wire-seq-no*)
;
;   If the name ends in a number already, we add a colon and suffix our
;   own number.  We also handle any users that do 'W when maybe they
;   shouldn't.
;
(define (check-wire-name foo)
   (when (and (pair? foo) (eq? (car foo) 'quote))
      (set! foo (cadr foo)))
   (let ((lc (char->integer
                (car (reverse (string->list (symbol->string foo)))))))
      (if (and (<= (char->integer #\0) lc)
               (<= lc (char->integer #\9)))
         (cons-symbol foo ':)
         foo)))
;
(define (with-wires args global?)
   (if (eq? (car args) 'global)
      (with-wires (cdr args) #t)
      (let ((seq (if global? 'none (next-wire-seq))))
         (cons 'let (cons (map (lambda (a)
                                  (if (symbol? a)
                                     (list a (list 'make-wire '() a seq))
                                     (list (car a) (list 'make-wire
                                                      (cadr a) (car a) seq))))
                                (car args))
                             (cdr args))))))
;
(define (with-places args)
   (cons 'let
      (cons (map (lambda (a)
		      (if (symbol? a)
		         (list a '(make-place))
			 (list (car a) (cons 'make-place (cdr a)))))
	       (car args))
         (cdr args))))
;
(define (with-transitions args)
   (cons 'let
      (cons (map (lambda (a)
		      (if (symbol? a)
		         (list a '(make-transition))
			 (list (car a) (cons 'make-transition (cdr a)))))
	       (car args))
         (cdr args))))
;
(define (make-wire args asc)
   (let* ((init (if (safe-car args)
                   (simple-expr (car args) asc)
                   'unknown))
          (t (safe-caddr args))
          (t2 (safe-cadr args))
          (number (if t
                     (if (eq? t 'none) "" (simple-expr t asc))
                     (next-wire-seq)))
          (name (if t2 (if (eq? t 'none) t2 (check-wire-name t2)) 'W))
          (n (cons-symbol name number))
          (old (assq n asc))
          (new (list 'wire n init)))
      (if old
         (begin
            (when (number? init)
               (wire-init-force! (cdr old) init))
            (cdr old))
         (begin
            (add-to-environment (cons n new))
            new))))
;
(define (wire-name a)
   (cadr a))
;
;   No relation to the above; the following are for value wires.
;
(define *verify-df-wires?* #f)
;
(define (make-wire-read p w e)
   (let* ((np (new-place))
          (npm (new-place))
          (ok-inv (new-trans-from-place npm))
          (inv-eps (new-eps-place))
          (die (new-trans-from-place inv-eps))
          (dests (map edge-outnode-lambda e))
          (dest-posn (map edge-outpos-lambda e)))
      (node-marked-set! npm #t)
      (connect ok-inv npm)
      (connect inv-eps ok-inv)
      (connect (vwire-invalidt w) inv-eps)
      (connect np die)
      (new-place-from-trans die)
      (for-each (lambda (a) (remove-edge a)) e)
      (let loop ((el (node-inedges p)))
         (when el
            (let* ((t (edge-innode (car el)))
                   (r (new-reg-from t (vwire-join w))))
               (for-each (lambda (a b) (connect-df r a b))
                  dests dest-posn)
               (connect t np)
               (connect npm t)
               (connect t (vwire-begread w)))
            (loop (cdr el))))
      (let loop ((el (node-outedges p)))
         (when el
            (let ((t (edge-outnode (car el))))
               (connect np t)
               (connect t npm))
            (loop (cdr el))))))
;
(define (make-wire-write p w e)
   (let* ((src (edge-innode (car e))))
      (remove-edge (car e))
      (let loop ((el (node-inedges p)))
      (when el
            (let* ((t (edge-innode (car el)))
                   (r (new-reg-from t src)))
               (connect-df r (vwire-join w) 0)
               (connect t (vwire-begwrite w)))
            (loop (cdr el))))
      (let loop ((el (node-outedges p)))
         (when el
            (let* ((t (edge-outnode (car el)))
                   (r (new-reg-from t (new-func '?))))
               (connect-df r (vwire-join w) 0)
               (connect t (vwire-endwrite w)))
            (loop (cdr el))))))
;
;   Finds a wire name.  The only one will be the one that doesn't
;   start with ':'.  There had better only be one.
;
(define (find-wire-name p)
   (let loop ((lbls (node-label p)) (sofar '()))
      (if lbls
         (if (eq? (car (string->list (symbol->string (car lbls)))) #\[)
            (loop (cdr lbls) sofar)
            (loop (cdr lbls) (cons (car lbls) sofar)))
         (if (= (length sofar) 1)
            (begin
               (node-label-set! p (remove (car sofar) (node-label p)))
               (car sofar))
            (merror "Needed exactly one wire label here" sofar)))))
;
;   This function runs through the places, checking each for
;   input or output data flow values.  When such are found, it
;   hooks up the wires, as necessary.
;
(define (connect-value-wires)
   (start-remove-cache)
   (let loop ((pl (graph-nodes *graph*)) (wasc '()))
      (if pl
         (let ((p (car pl)))
            (when (node-type? p 'P)
               (let ((inedge (remove-if (lambda (a)
                                            (not (eq? (edge-type a) 'D)))
                                (node-inedges p)))
                     (outedge (remove-if (lambda (a)
                                            (not (eq? (edge-type a) 'D)))
                                 (node-outedges p))))
                  (when (or inedge outedge)
                     (when (and inedge outedge)
                        (merror "Input *and* output vwires from place??" 0))
                     (let* ((wn (find-wire-name p))
                            (w (assq wn wasc)))
                        (when (not w)
                           (set! wasc (cons (cons wn (new-vwire wn)) wasc))
                           (set! w (assq wn wasc)))
                        (if inedge
                           (make-wire-write p (cdr w) inedge)
                           (make-wire-read p (cdr w) outedge))))))
            (loop (cdr pl) wasc))
         (clean-remove-cache))))
;
;   We only allow a merge here if p1 is un-annotated and if it has no
;   outputs.  Otherwise, we introduce a transition.
;
(define (virgin-place? p)
   (and (= (length (node-outedges p)) 0)
        (null? (node-label p))
        (not (node-marked p))))
;
(define (virgin-place-notime? p)
   (and (= (length (node-outedges p)) 0)
;       (null? (node-label p))
        (not (node-marked p))))
;
;   If the only output edges are data flow then we go ahead and treat it as
;   though it were virgin.
;
(define (just-df-stuff p)
   (and (not (node-marked p))
      (let loop ((el (node-outedges p)))
         (or (null? el)
	    (eq? (edge-type a) 'P)
	    (loop (cdr el))))))
;
(define (virgin-place p)
   (if (virgin-place? p) p
      (new-place-from-trans (new-trans-from-place p))))
;
(define (temp-violate p)
   (if (virgin-place? p)
      (add-to-label! p 'takeaway))
   p)
;
(define (unviolate p)
   (node-label-set! p (remove-if (lambda (a) (eq? a 'takeaway))
                         (node-label p))))
;
(define (try-merge-nodes p1 p2)
   (merge-nodes p1 (virgin-place p2)))
;
(define (merge-nodes t1 t2)
   (if (not (eq? t1 t2))
      (let ((inedges (node-inedges t2))
            (outedges (node-outedges t2)))
         (let loop ((el inedges))
            (if el (begin
                      (edge-outnode-set! (car el) t1)
                      (loop (cdr el)))))
         (let loop ((el outedges))
            (if el (begin
                      (edge-innode-set! (car el) t1)
                      (loop (cdr el)))))
         (node-inedges-set! t1 (append inedges (node-inedges t1)))
         (node-outedges-set! t1 (append outedges (node-outedges t1)))
         (node-inedges-set! t2 '())
         (node-outedges-set! t2 '())
;
;   If it is a trans, we patch up graph-t-assoc.  This is the only
;   place that we substitute one transition for another.
;
         (if (node-type? t1 'T)
            (let loop ((labels (reverse (node-label t2))))
               (if labels
                  (begin
                     (if (assoc (caar labels) (node-label t1))
;
;   If we this particular label is the same as one on the transition
;   we are merging to, we just try to pop it off.
;
                        (let ((a (assoc (caar labels)
                                    (graph-t-assoc *graph*))))
                           (set-cdr! a
                              (remove-if (lambda (a) (eq? (car a) t2))
                                 (cdr a))))
                        (begin
                           (let loop ((v (cdr (assoc (caar labels)
                                                 (graph-t-assoc *graph*)))))
                              (when v
				 (if (eq? (caar v) t2)
				    (set-car! (car v) t1))
			         (loop (cdr v))))
                           (add-to-label! t1 (car labels))))
                     (loop (cdr labels)))))
            (let loop ((labels (reverse (node-label t2))))
               (when labels
		  (add-to-label! t1 (car labels))
	          (loop (cdr labels)))))
         (remove-node t2))))
;
;  Next, our routines to simplify keys.
;
(define (find-input-node-of-type t p)
   (let ((res (remove-if (lambda (a) (not (node-type? a t)))
                 (node-in-nodes p))))
      (if (= (length res) 1)
         (car res)
         (merror "More than one such node into this node" 0))))
;
(define (find-output-node-of-type t p)
   (let ((res (remove-if (lambda (a) (not (node-type? a t)))
                 (node-out-nodes p))))
      (if (= (length res) 1)
         (car res)
         (merror "More than one such node out of this node" p))))
;
;   This function `clones' a transition and any associated registers.
;
(define (clone-transition orig)
   (let ((new (new-trans)))
      (node-label-set! new (node-label orig))
      (let loop ((el (node-inedges orig)))
         (when el
            (let ((n (edge-innode (car el))))
               (if (eq? (edge-type (car el)) 'D)
                  (connect-df n new (edge-outpos (car el)))
                  (when (node-type n)
                     (edge-ignore-timing?-set! (connect n new)
                        (edge-ignore-timing? (car el)))))
               (loop (cdr el)))))
      (let loop ((el (node-outedges orig)))
         (when el
            (let ((n (edge-outnode (car el))))
               (if (node-type? n 'R)
                  (let ((newr (new-reg-from new
                                 (edge-innode
                                    (if (node-type?
                                           (edge-innode
                                              (car (node-inedges n))) 'T)
                                       (cadr (node-inedges n))
                                       (car (node-inedges n)))))))
                     (connect-df newr
                        (edge-outnode (car (node-outedges n))) 0))
                  (when (node-type n)
                     (connect new n)))
               (loop (cdr el)))))
      new))
;
;   Removes a transition and any associated registers.
;
(define (remove-node-t orig)
    (let loop ((el (node-outedges orig)))
       (when el
          (let ((n (edge-outnode (car el))))
             (when (node-type? n 'R)
	        (remove-node n)))
          (loop (cdr el))))
    (remove-node orig))
;
(define (fix-simplify-key k)
   (let* ((p (cadr k))
          (j (if (cddr k) (caddr k) '()))
          (inarity (length (node-inedges p)))
          (outarity (length (node-outedges p))))
;      (if (or (= inarity 0) (= outarity 0))
;         (merror "Oops, key with no input or output" k))
;
;   Now we fix any places that lead to this place, or places that this place
;   leads to, that were created when the preceding transition didn't exist.
;
      (let loop ((edges (node-inedges p)))
         (if edges
            (let ((e (car edges)))
               (when (node-type? (edge-innode e) 'P)
                  (move-edge-head e (edge-innode e)
                     (find-input-node-of-type 'T (edge-innode e))))
               (loop (cdr edges)))))
      (let loop ((edges (node-outedges p)))
         (if edges
            (let ((e (car edges)))
               (when (node-type? (edge-outnode e) 'P)
                  (move-edge-tail e (edge-outnode e)
                     (find-input-node-of-type 'T (edge-outnode e))))
               (loop (cdr edges)))))
      (when (and (= inarity 1) (= outarity 1))
         (let ((t1 (edge-innode (car (node-inedges p))))
                (t2 (edge-outnode (car (node-outedges p)))))
             (remove-node p)
             (merge-nodes t1 t2)))
;            (merge-nodes (edge-innode (car (node-inedges p)))
;                         (edge-outnode (car (node-outedges p))))
;            (remove-node p)))
      (if j
         (let* ((inreg
                   (if (> inarity 0) (edge-innode (car (node-inedges j)))))
                (outreg
                   (if (> outarity 0) (edge-outnode (car (node-outedges j))))))
            (if (and j (= inarity 1) (node-type? inreg 'R))
               (let* ((up (node-in-nodes inreg))
                      (dnpos (map edge-outpos-lambda (node-outedges j)))
                      (dn (node-out-nodes j)))
                  (remove-node j)
                  (remove-node inreg)
                  (map (lambda (outreg)
                           (if (not (node-type? outreg 'T))
                              (map (lambda (d w) (connect-df outreg d w))
                                 dn dnpos)))
                     up))
               (if (and j (= outarity 1) (node-type? outreg 'R))
                  (let* ((dn (node-out-nodes outreg))
                         (up (node-in-nodes j))
                         (dnpos
                            (map edge-outpos-lambda (node-outedges outreg))))
                     (remove-node j)
                     (remove-node outreg)
                     (map (lambda (a)
                             (map (lambda (b w) (connect-df a b w)) dn dnpos))
                        up))
                  (if (and j (= inarity 1) (= outarity 1))
                     (let ((innode (edge-innode (car (node-inedges j))))
                           (outnode (edge-outnode (car (node-outedges j))))
                           (outpos (edge-outpos (car (node-outedges j)))))
                        (remove-node j)
                        (connect-df innode outnode outpos)))))))))
;
(define (fix-simplify-keys)
   (let loop ((keys (graph-keys *graph*)))
      (if keys
         (begin
            (fix-simplify-key (car keys))
            (loop (cdr keys))))))
;
(define *expand-eps-places* #f)
;
(define (is-eps-place? p)
   (and (node-type? p 'P)
                     (= 1 (length (node-label p)))
                     (or
                        (eq? *eps-brackets* (car (node-label p)))
                        (eq? *eps-immediate* (car (node-label p))))))
;
(define (is-simple-eps-place? p)
   (and (node-type? p 'P)
                     (= 1 (length (node-label p)))
                     (eq? *eps-brackets* (car (node-label p)))))
;
(define (one-in-out? p)
   (= 1 (length (node-inedges p)) (length (node-outedges p))))
;
(define (no-common-in-places t1 t2)
   (let ((check (intersection (map edge-innode-lambda (node-inedges t1))
                              (map edge-innode-lambda (node-inedges t2)))))
      (not (and check
              (remove-if (lambda (a) (not (node-type? a 'P))) check)))))
;
;   If the epsilon place is marked, we squeeze the marking down to the
;   next place.  If there are multiple successor transitions, we
;   complain.
;
(define (remove-extra-epsilon-places)
   (start-remove-cache)
   (let loop ((places (graph-nodes *graph*)))
      (when places
         (let ((p (car places)))
            (if (and (is-simple-eps-place? p)
                     (or *expand-eps-places* (one-in-out? p)))
               (let ((in (map edge-innode-lambda (node-inedges p)))
                     (out (map edge-outnode-lambda (node-outedges p))))
                  (when (node-input-output? p)
                     (for-each (lambda (p1)
                                  (when (node-type? p1 'T)
                                     (for-each (lambda (e)
                                                  (let ((n (edge-innode e)))
                                                     (when (node-type? n 'P)
                        (node-input-output?-set! n (node-input-output? p)))))
                                        (node-inedges p1))))
                        in))
                  (when (node-marked p)
                     (when (not (= 1 (length out)))
                        (merror "Too many nodes out of marked eps place" 0))
                     (when (node-label (car out))
                        (merror "Can't squeeze marking past labeled transition"
                           (node-label (car out))))
                     (let loop ((pl (map edge-outnode-lambda
                                       (node-outedges (car out)))))
                        (when pl
                           (begin
                              (node-marked-set! (car pl) #t)
                              (loop (cdr pl))))))
                  (remove-node p)
                  (let loop ((tl1 in))
                     (when tl1
                        (let loop ((tl2 out))
                           (when tl2
                              (when (no-common-in-places (car tl1) (car tl2))
; (display-stuff "Trying to merge two transitions . . ." 'newline)
                                 (merge-nodes (clone-transition (car tl1))
                                              (clone-transition (car tl2))))
                              (loop (cdr tl2))))
                        (loop (cdr tl1))))
                  (for-each remove-node-t in)
                  (for-each remove-node-t out))
               (when (and (node-type? p 'T)
                        (null? (node-inedges p))
                        (= 1 (length (node-outedges p))))
                  (let* ((p2 (edge-outnode (car (node-outedges p))))
                         (t2 (edge-outnode (car (node-outedges p2)))))
                     (when (and (eq? *eps-immediate* (car (node-label p2)))
                              (null? (node-outedges t2))
                              (= 1 (length (node-inedges t2))))
                        (remove-node p2)
                        (remove-node p)
                        (remove-node t2))))))
            (loop (cdr places))))
   (clean-remove-cache))
;
(define (iset-member? h key)
   (let loop ((hc (vector-ref (hash-table h) (modulo key (hash-prime h)))))
      (and hc (if (eqv? key ((hash-keyfunc h) (car hc)))
                 (car hc) (loop (cdr hc))))))
;
;   For this one, not only the key but the entire entry must match.  This
;   is useful for arrays, for instance.
;
(define (iset-member*? h v)
   (let ((key ((hash-keyfunc h) v)))
      (let loop ((hc (vector-ref (hash-table h) (modulo key (hash-prime h)))))
         (and hc (if (equal? v (car hc)) (car hc) (loop (cdr hc)))))))
;
;   Here, only the car of the entry must match.
;
(define (iset-member**? h v)
   (let ((key ((hash-keyfunc h) (list v))))
      (let loop ((hc (vector-ref (hash-table h) (modulo key (hash-prime h)))))
         (and hc (if (equal? v (caar hc)) (car hc) (loop (cdr hc)))))))
;
;   This function finds the next prime number greater than or equal to n,
;   assuming n is odd.
;
(define (next-prime n)
   (let loop ((d 3))
      (if (> (* d d) n) n
         (if (= (modulo n d) 0) (next-prime (+ n 2))
            (loop (+ d 2))))))
;
;   This allocates a new hash table.
;
(define (new-hash func)
   (new-hash-aux 353 func))
;
(define (new-hash-aux n func)
   (let ((h (make-hash)))
      (hash-prime-set! h n)
      (hash-limit-set! h (if *small-and-slow?* (* 3 n) n))
      (hash-size-set! h 0)
      (hash-table-set! h (make-vector n '()))
      (hash-keyfunc-set! h func)
      h))
;
;   This is a generic hash iteration function.
;
(define (hash-iter func h)
   (let loop ((i 0))
      (when (< i (hash-prime h))
         (for-each func (vector-ref (hash-table h) i))
         (loop (+ i 1)))))
;
;   This rehashes when we get too full.  Grows more slowly if we are in
;   slow mode (1.5 vs 3).
;
(define (rehash h)
   (let* ((nprime (next-prime
		     (if *small-and-slow?*
                        (+ 200 (quotient (* 3 (hash-prime h)) 2))
                        (+ 200 (* 3 (hash-prime h))))))
          (nh (new-hash-aux nprime (hash-keyfunc h))))
      (when *small-and-slow?*
         (display-stuff "{->" nprime "}"))
      (let loop ((i 0))
         (when (< i (hash-prime h))
	    (for-each (lambda (a) (iset-insert nh a))
	       (vector-ref (hash-table h) i))
	    (vector-set! (hash-table h) i '())
	    (loop (+ i 1))))
      (hash-prime-set! h nprime)
      (hash-limit-set! h (if *small-and-slow?* (* 3 nprime) nprime))
      (hash-table-set! h (hash-table nh))))
;
;   This function does an unconditional insert.
;
(define (iset-insert h k)
   (when (>= (hash-size h) (hash-limit h))
      (rehash h))
   (let ((hv (modulo ((hash-keyfunc h) k) (hash-prime h))))
      (hash-size-set! h (+ 1 (hash-size h)))
      (vector-set! (hash-table h) hv (cons k (vector-ref (hash-table h) hv)))
      k))
;
;   This function takes the orbits, numbered 0..n, with gaps, and removes
;   the gaps.  It does this with an array containing a list of all of the
;   places that have attached orbits, and then scanning the array,
;   renumbering the orbits.
;
;   (This is somewhat dangerous; the maxorbit can be a very large number
;   and then we construct here a potentially very large array.)  FIXME
;
;   Okay, this has become important.  We insert the hash stuff here and
;   use it.
;
(define (minimize-orbits)
   (let ((h (new-hash car)) (i 0))
      (for-each
         (lambda (p)
            (when (and (node-type? p 'P) (node-orbit p))
               (let* ((o (node-orbit p))
                      (v (or (iset-member? h o) (iset-insert h (list o)))))
                  (set-cdr! v (cons p (cdr v))))))
         (graph-nodes *graph*))
      (hash-iter
         (lambda (p)
            (for-each (lambda (p) (node-orbit-set! p i)) (cdr p))
            (set! i (+ i 1)))
         h)
      (graph-maxorbits-set! *graph* i)
      (when *expand-eps-places* (make-sgraph-fsm))))
;
;   Then, if we are expanding epsilon places, we also add in the additional
;   tokens required to make the graph a network of FSMs.  We do this by
;   scanning the places, keeping track of which places are marked.  We
;   then scan the transitions, and any transitions that do not set the
;   same set of input/output orbits gets annotated with additional
;   edges to or from the appropriate new place.
;
(define (make-sgraph-fsm)
   (let ((pn (make-vector (graph-maxorbits *graph*) '())))
      (for-each (lambda (p)
                   (when (and (node-type? p 'P)
                              (node-marked p)
                              (not (is-eps-place? p)))
                      (vector-set! pn (node-orbit p) p)))
         (graph-nodes *graph*))
      (let loop ((i 0))
         (when (< i (graph-maxorbits *graph*))
            (set! *current-orbit* i)
            (vector-set! pn i (if (vector-ref pn i) '()
                                 (let ((p (new-place)))
                                    (node-marked-set! p #t)
                                    p)))
            (loop (+ i 1))))
      (for-each (lambda (t)
                   (when (node-type? t 'T)
                      (let ((in (remove-if not
                                   (map (lambda (a)
                                           (node-orbit (edge-innode a)))
                                      (node-inedges t))))
                            (out (remove-if not
                                    (map (lambda (a)
                                            (node-orbit (edge-outnode a)))
                                       (node-outedges t)))))
                         (for-each (lambda (p)
                                      (when (not (member p out))
                                         (connect t (vector-ref pn p))))
                            in)
                         (for-each (lambda (p)
                                      (when (not (member p in))
                                         (connect (vector-ref pn p) t)))
                            out))))
         (graph-nodes *graph*))))
;
;   Now we have chosen a generator and a bunch of receivers.  We
;   connect them up!
;
(define (connect-gen-rec gen recs)
   (if (= (length recs) 1)
      (let ((p (new-eps-place)))
         (let loop ((ts gen))
            (when ts
	       (connect (car ts) p)
	       (loop (cdr ts))))
         (let loop ((ts (car recs)))
            (when ts
	       (connect p (car ts))
	       (loop (cdr ts)))))
      (begin
         (let ((t (if (= (length gen) 1) (car gen)
                     (let* ((np (new-eps-place))
                            (nt (new-trans-from-place np)))
                        (let loop ((ts gen))
			   (when ts
                              (connect (car ts) np)
                              (loop (cdr ts))))
                        nt))))
            (let loop ((recs recs))
               (if recs
                  (let ((np (new-place-from-trans t)))
                     (add-to-label! np *eps-brackets*)
                     (let loop ((ts (car recs)))
                        (when ts
			   (connect np (car ts))
			   (loop (cdr ts))))
                     (loop (cdr recs)))))))))
;
;   When we hit this function, we have the set of groups divided up into
;   those with just one element and those with more than one element.
;   We start by joining all the one elements together.
;
(define (connect-ones-twos ones twos)
; (display-stuff "% connect of " event " has " (length ones) " ones and "
;      (length twos) " twos." 'newline)
   (when (> (length ones) 1)
      (let loop ((t (caar ones)) (rest (cdr ones)))
         (when rest
	    (merge-nodes t (caar rest))
            (loop t (cdr rest)))))
   (if ones
      (connect-gen-rec (car ones) twos)
      (connect-gen-rec (car twos) (cdr twos))))
;
;   This function handles connecting a bunch of like transitions together.
;   We segregate the transitions into their respective groups.
;
(define (connectemall trans)
   (let ((gv (make-vector (graph-maxgroupval *graph*) '())))
      (let loop ((tlist trans))
         (when tlist
	    (vector-set! gv (cdar tlist)
               (cons (caar tlist) (vector-ref gv (cdar tlist))))
            (loop (cdr tlist))))
      (let ((groups (remove-if null? (vector->list gv))))
         (if (> (length groups) 1)
            (let loop ((rest groups) (ones '()) (twos '()))
               (if rest
                  (if (= (length (car rest)) 1)
                     (loop (cdr rest) (cons (car rest) ones) twos)
                     (loop (cdr rest) ones (cons (car rest) twos)))
                  (connect-ones-twos ones twos)))))))
;
;   Here we go through the entire graph, unifying groups of transitions.
;   This is optional.
;
(define (connect-groups)
   (start-remove-cache)
   (let loop ((events (graph-t-assoc *graph*)))
      (if events
         (let* ((trans (cdar events)))
            (connectemall trans)
            (loop (cdr events)))
         (clean-remove-cache))))
;
(define (single-label t)
   (if (or (symbol? t) (number? t)) t
      (if (and (pair? t) (number? (cdr t)))
         (single-label (car t))
         (apply cons-symbol t))))
;
(define *show-unfired-transitions?* #f)
;
(define (printable-label-list labels)
   (let loop ((sofar '()) (labels labels))
      (if labels
         (if (eq? (car labels) 'takeaway)
            (loop sofar (cdr labels))
            (loop (if sofar (cons-symbol sofar "," (single-label (car labels)))
                     (cons-symbol (single-label (car labels))))
                  (cdr labels)))
         (if sofar sofar ""))))
;
(define (label-list t)
   (let loop ((sofar '()) (labels (node-label t)))
      (if labels
         (if (eq? (car labels) 'takeaway)
            (loop sofar (cdr labels))
            (loop (cons (single-label (car labels)) sofar) (cdr labels)))
         sofar)))
;
(define (printable-label n)
   (let ((nl (node-label n)))
      (if (and (node-type? n 'P) nl (symbol? (car nl))
             (let ((s (symbol->string (car nl))))
                (and (eq? #\[ (string-ref s 0))
                   (not (or (string=? s "[]") (string=? s "[!]"))))))
         (cons-symbol (printable-label-list nl)
            (if (eq? (node-input-output? n) '*input*) 'c 'b))
         (printable-label-list nl))))
;
;   If this flag is set true, we ignore any portion of the label that
;   contains an at sign (as intermediate wires do.)
;
(define *use-shorter-labels?* #t)
;
(define (ignorable-label l)
   (and *use-shorter-labels?* (symbol? l)
      (memq #\@ (string->list (symbol->string l)))))
;
;   This tortured function is because Elk doesn't always like nested
;   named lets.
;
(define (get-more-label ls sofar)
   (if ls
      (let ((l (single-label (car ls))))
         (if (or (member l sofar) (ignorable-label l))
            (get-more-label (cdr ls) sofar)
            (get-more-label (cdr ls) (cons l sofar))))
      sofar))
;
(define (printable-list-label tl)
   (let loop ((tl tl) (sofar '()))
      (if tl
         (loop (cdr tl) (get-more-label (node-label (car tl)) sofar))
         (printable-label-list sofar))))
;
(define *use-sim-state* #f)
;
(define (print-node-type n)
   (let ((p (node-type n)))
      (if (and (eq? p 'P) 
               (if *use-sim-state*
		  (sim-marked? (node-snode n))
		  (node-marked n)))
	 'M
	 (if (and (eq? p 't) *show-unfired-transitions?* (node-snode n)
                (not (strans-fired?-lambda (node-snode n))))
	    'U
	    p))))
;
(define (node-in-nodes n)
   (map edge-innode-lambda (node-inedges n)))
;
(define (node-out-nodes n)
   (map edge-outnode-lambda (node-outedges n)))
;
;   This function indicates whether the node is either a place of the
;   appropriate orbit or another node connected to a place of the
;   appropriate orbit.
;
(define (node-in-orbit? n o)
   (if (node-type? n 'P)
      (eqv? o (node-orbit n))
      (let loop ((nl (append (node-in-nodes n) (node-out-nodes n))))
         (if nl
            (let ((n (car nl)))
               (or (and (node-type? n 'P) (eqv? o (node-orbit n)))
                   (loop (cdr nl))))
            #f))))
;
;   This iterates over the graph edges.
;
(define (ge-for-each func)
   (for-each (lambda (n)
                (for-each func (node-outedges n))) (graph-nodes *graph*)))
;
;   This function sends a graph such that all nodes fit a particular
;   predicate.  If the thing passed is a file, then it doesn't open
;   or close it.
;
(define (send-subgraph f g which)
   (let ((nf (if (string? f) (open-output-file f) f)))
      (for-each (lambda (n)
                   (when (which n)
                      (display-stuff-port nf "N " (print-node-type n) " "
                         (node-name n) " " (printable-label n) 'newline)))
         (graph-nodes g))
      (ge-for-each (lambda (e)
                   (let ((n1 (edge-innode e)) (n2 (edge-outnode e)))
                      (when (and (which n1) (which n2))
                         (display-stuff-port nf "E " (edge-type e)
                            (if (edge-ignore-timing? e) "e " " ")
                            (node-name n1) " " (edge-inpos e) " "
                            (node-name n2) " " (edge-outpos e) " "
			    'newline)))))
      (when (string? f) (close-output-port nf))))
;
(define (send-graph f g)
   (send-subgraph f g (lambda (ignore) #t)))
;
(define (send-subgraph-by-orbits f g)
   (let ((f (open-output-file f)))
      (let loop ((i 0))
         (if (< i (graph-maxorbits g))
            (begin
               (send-subgraph f g (lambda (a) (node-in-orbit? a i)))
               (display-stuff-port f "ZZZ" 'newline)
               (loop (+ i 1)))
            (close-output-port f)))))
;
(define (test f e)
   (display-stuff f 'newline)
   (send-graph f (build-graph e)))
;
(define (test-sep f e)
   (set! *connect-groups?* #f)
   (display-stuff f 'newline)
   (send-graph f (build-graph e))
   (set! *connect-groups?* #t))
;
;   Prints the current simulation state.
;
(define (curstate f)
   (display-stuff f 'newline)
   (when *sgraph*
      (set! *use-sim-state* #t))
   (send-graph f *graph*)
   (set! *use-sim-state* #f))
;
;   This is where we deal with subcircuits, which are defined using
;   our own home-grown macros.
;
(define (expr-macro? s)
   (and (pair? s) (eq? (car s) 'macro)))
;
(define (circuit? val)
   (and (expr-macro? val) (> (+ (macro-inputs val) (macro-outputs val)) 0)))
;
(define (verify-ckt? exp asc)
   (let ((val (simple-expr exp asc)))
      (if (circuit? val)
         val
         (merror "Not a circuit!" exp))))
;
;   This function has the added side effect of writing out the schematic
;   if this option has been enabled.
;
(define *write-schematic-file* #f)
;
(define *schematic-wires-seen* '())
;
(define *schematic-outputs-handled* '())
;
(define *schematic-queue* '())
;
(define *schematic-counter* 0)
;
(define (write-schematic n)
   (set! *schematic-outputs-handled* '())
   (set! *schematic-counter* 0)
   (set! *schematic-queue* '())
   (set! *schematic-wires-seen* '())
   (set! *write-schematic-file* (open-output-file n)))
;
(define (schematic-wire w)
   (let ((v (assoc w *schematic-wires-seen*)))
      (if v (cdr v)
         (begin
            (set! *schematic-wires-seen*
               (cons (cons w (cons-symbol 'W *schematic-counter*))
                  *schematic-wires-seen*))
            (set! *schematic-counter* (+ 1 *schematic-counter*))
            (display-stuff-port *write-schematic-file*
               "N E " (cdar *schematic-wires-seen*) " " w 'newline)
            (schematic-wire w)))))
;
(define (schematic-edges wl f pos in?)
   (when wl
      (if in?
         (display-stuff-port *write-schematic-file*
            "E P " (car wl) " 0 " f " " pos 'newline)
         (display-stuff-port *write-schematic-file*
            "E P " f " " pos " " (car wl) " 0" 'newline))
      (schematic-edges (cdr wl) f (+ pos 1) in?)))
;
(define (schematic-node-write entity)
   (let ((ins (cadr entity))
         (outs (caddr entity))
         (name (car entity)))
      (when (null? (intersection outs *schematic-outputs-handled*))
         (set! *schematic-outputs-handled*
            (append outs *schematic-outputs-handled*))
         (let* ((nins (map schematic-wire ins))
                (nouts (map schematic-wire outs))
                (func (cons-symbol 'F *schematic-counter*)))
            (display-stuff-port *write-schematic-file*
               "N F " func " " name 'newline)
            (schematic-edges nins func 0 #t)
            (schematic-edges nouts func 0 #f)))))
;
(define (close-schematic)
   (for-each schematic-node-write *schematic-queue*)
   (close-output-port *write-schematic-file*)
   (set! *write-schematic-file* #f))
;
(define (schematic-node-save mac args)
   (let ((ins (my-sublist args 0 (length (cadr mac))))
         (outs (my-sublist args (length (cadr mac)) (length (caddr mac))))
         (name (cadr (cddr mac))))
      (when name
         (set! *schematic-queue* (cons (list name ins outs)
                                              *schematic-queue*)))))
;
;   Doing a macro is much the same as doing a procedure call, only instead
;   of 
;
(define (macro-name mac)
   (nth 1 mac))
;
(define (macro-name-set! mac v)
   (set-car! (cdr mac) v))
;
(define (macro-body mac)
   (nth 3 mac))
;
(define (macro-formals mac)
   (nth 2 mac))
;
(define (macro-inputs mac)
   (nth 4 mac))
;
(define (macro-outputs mac)
   (nth 5 mac))
;
(define (macro-mirror? mac)
   (and (> (length mac) 6) (eq? (nth 6 mac) '*mirror*)))
;
(define (wire-init-set! w nv)
   (let ((v (caddr w)))
      (when (and (number? v) (not (= v nv)))
         (merror "Wire already has a different value" w))
      (set-car! (cddr w) nv)))
;
(define (wire-init-force! w nv)
   (let ((v (caddr w)))
      (set-car! (cddr w) nv)))
;
;   Handles tracking what are inputs and what are outputs.  Ugly, but
;   functional.
;
(define (find-io-in-environment asc)
   (or (safe-cdr (assq '*IO* asc)) '()))
;
(define (mirror-input-output foo)
   (if (eq? foo '*output*) '*input*
      (if (eq? foo '*input*) '*output*
         foo)))
;
(define (check-io-in-environment w asc)
; (display-stuff "Checking " w " in " (find-io-in-environment asc) 'newline)
   (let ((io (find-io-in-environment asc)))
      (if (assq '*mirror* io)
         (mirror-input-output
            (safe-cdr (assoc w (find-io-in-environment asc))))
         (safe-cdr (assoc w (find-io-in-environment asc))))))
;
(define (extend-io mirror? args ins outs asc)
; (display-stuff "Extending IO with " args " (" ins "," outs ")" 'newline)
   (cons (cons '*IO*
            (append (map (lambda (a) (cons a '*input*)) (my-sublist args 0 ins))
                    (map (lambda (a) (cons a '*output*))
                       (my-sublist args ins outs))
                    (if mirror?
                       (cons (list '*mirror*) (find-io-in-environment asc))
                       (find-io-in-environment asc))))
      asc))
;
;   We want to implement static scoping, but if we do, handling the
;   dynamic scoping required by completeness stuff will be difficult.
;   So we just use dynamic scoping.  This will only make a difference
;   if the user makes an error (since we don't tell him that we have
;   dynamic scoping), in which case the expression will be accepted
;   when it shouldn't be.  Perhaps later we'll improve this.
;
;   We also keep track of what are inputs and what are outputs in this
;   function by adding something of the form *IO* to the global
;   association list.
;
(define (do-macro mac args place asc)
   (let* ((pars (build-expr-par args place asc))
          (nargs (map car (cdr pars))))
      (when (not (= (length nargs) (length (macro-formals mac))))
         (merror "Bad number of arguments to " (macro-name mac)))
      (when (> (+ (macro-inputs mac) (macro-outputs mac)) 0)
         (set! asc
            (extend-io (macro-mirror? mac) nargs
               (macro-inputs mac) (macro-outputs mac) asc)))
      (build-expr (macro-body mac) (car pars)
         (append (map cons (macro-formals mac) nargs)
            asc))))
;
;   This is what the last line above used to be.
;
;           (list (assq 'GLOBAL asc)) *global-assoc*
;
(define (my-list? a)
   (or (pair? a) (null? a)))
;
(define (check-all-symbols lst)
   (when lst
      (if (not (symbol? (car lst)))
         (merror "Got a non-symbol in macro def" (car lst))
         (check-all-symbols (cdr lst))))
   lst)
;
(define (macro-filter-header header)
   (if (and (>= (length header) 3)
            (my-list? (cadr header)) (my-list? (caddr header)))
      (let ((in (cadr header)) (out (caddr header)))
         (list (check-all-symbols (append in out (cdddr header)))
            (length in) (length out)))
      (list (check-all-symbols (cdr header)) 0 0)))
;
(defmac (def pattern . body)
   `(add-to-environment
      ,(if (symbol? pattern)
         (let ((r (simple-expr (cons 'begin body) *global-assoc*)))
            (when (and (expr-macro? r) (null? (macro-name r)))
               (macro-name-set! r pattern))
            `(cons ',pattern ',r))
         (let ((fa (macro-filter-header pattern)))
            `(cons ',(car pattern)
               (list 'macro ',(car pattern) ',(car fa)
                  ',(cons 'begin body) ',(cadr fa) ',(caddr fa)))))))
;
(defmac (verify exp)
   `(vtest ',exp '() '()))
;
(defmac (verify-bf exp)
   `(vtest-bf ',exp '() '()))
;
(defmac (verify-no-bf exp)
   `(vtest-no-bf ',exp '() '()))
;
(defmac (simulate exp)
   `(stest ',exp '() '()))
;
(defmac (save-graph . args)
   `(save-graph-internal ',args))
;
(defmac (save-graph-current-state . args)
   `(save-graph-cs-internal ',args))
;
(defmac (save-state-graph . args)
   `(save-state-graph-internal ',args))
;
(defmac (X-display-graph . args)
   `(X-display-graph-internal ',args))
;
(defmac (PS-display-graph . args)
   `(PS-display-graph-internal ',args))
;
(defmac (graph-to-ps . args)
   `(graph-to-ps-internal ',args))
;
(define (split-state state arity)
   (if (<= arity 0) '()
      (cons (remainder state 2) (split-state (quotient state 2) (- arity 1)))))
;
;   We cheat with pow2; we cache results with an array to make things
;   somewhat faster.  (If only we could do the same thing with the
;   logical or and and functions, which scheme notably lacks!)
;
(define *pow2-array* '())
;
(define *pow2-limit* 0)
;
(define (fill-pow2 n)
   (set! *pow2-array* (make-vector n '()))
   (set! *pow2-limit* n)
   (vector-set! *pow2-array* 0 1)
   (let loop ((i 1))
      (when (< i n)
         (vector-set! *pow2-array* i (* 2 (vector-ref *pow2-array* (- i 1))))
         (loop (+ i 1)))))
;
(define (pow2 n)
   (when (>= n *pow2-limit*)
      (fill-pow2 (* n 2)))
   (vector-ref *pow2-array* n))
;
(pow2 100)
;
(define (unsplit-state bits)
   (if bits (+ (car bits) (* 2 (unsplit-state (cdr bits)))) 0))
;
;   A combinational function takes a set of inputs, and the
;   output is automatically calculated.  If it is supplied, then
;   it is checked.
;
;   Usage:  (sm-ckt (/nor/ (a b) (c))
;              (lambda (a b) (and a b)))
;
(defmac (sm-ckt formals func)
   `(sm-ckt-really ',formals ',func))
;
(defmac (global-wires . body)
   (list 'build-expr `',(cons 'global-wires-def body)
      '(new-place-unadded) '*global-assoc*))
;
(define (map-to-wire-inits expr what)
   (if (my-list? expr)
      (map (lambda (a) (map-to-wire-inits a what)) expr)
      (if (memq expr what)
         `(wire-init ,expr)
         expr)))
;
(define (sm-expand-init ins)
   (if (cdr ins)
      `(+ (wire-init ,(car ins)) (* 2 ,(sm-expand-init (cdr ins))))
      `(wire-init ,(car ins))))
;
(define (sm-apply fake-macro args place)
   (let ((r (do-macro fake-macro args place *global-assoc*)))
      (if (eq? (car r) place)
         (cadr r)
         (merror "Something internal wrong in sm-ckt" args))))
;
(define (do-preeval fake-macro arity p)
   (let* ((n (pow2 arity)) (v (make-vector n 0)))
      (let loop ((i 0))
	 (when (< i n)
            (vector-set! v i (sm-apply fake-macro (split-state i arity) p))
	    (loop (+ i 1))))
      v))
;
;   This little hack changes
;
;   (case s
;      (v ...
;      (v ...
;
;   to
;
;   (case s #(...#))
;
;   It replaces the list with a vector that can be indexed quickly!
;   This should help things a lot.
;
(define (fix-case expr)
   (let ((n 0))
      (let loop ((l (cddr expr)))
         (when l
	    (when (>= (caar l) n)
	       (set! n (+ (caar l) 1)))
	    (loop (cdr l))))
      (let ((v (make-vector n '())))
         (for-each (lambda (l)
	              (vector-set! v (car l)
	                 (cons l (vector-ref v (car l)))))
	    (cddr expr))
	 (list 'case 's v))))
;
;   We used to use a lambda expression, and this was fairly quick.  Now
;   we use our `language' to interpret it, using macro calls.  This might
;   be slower, but at least it is more consistent.
;
(define (sm-ckt-really formals func)
   (let* ((name (car formals))
          (ins (cadr formals))
          (outs (caddr formals))
          (arity (length ins))
          (fake-macro (list 'macro 'foo ins func 0 0))
          (p (new-place-unadded))
	  (preeval (do-preeval fake-macro arity p)))
      (add-to-environment
         (cons name
            (list 'macro name (append ins outs)
               (list 'complete '()
                  `(wire-init-set! ,(car outs)
                     ,(map-to-wire-inits func ins))
                  (let loop ((state (- (pow2 arity) 1)) (sofar '()))
                     (if (>= state 0)
                        (let* ((args (split-state state arity))
                               (init-val (vector-ref preeval state)))
                           (let loop2 ((i 0) (b 1) (sofar sofar)
				       (niargs args) (niins ins))
                              (if (< i arity)
                                 (let* ((newstate
                                           ((if (= 0 (car niargs)) + -)
                                              state b))
                                        (newval (vector-ref preeval newstate)))
                                    (loop2 (+ i 1) (* b 2)
                                       (cons
                                          (if (= newval init-val)
                                             `(,state (annotate
                                               ,(cons-symbol (car niins)
                                                   (if (= 0 (car niargs))
                                                      '+ '-)))
                                               (loop ,newstate))
                                             `(,state (annotate
                                               ,(cons-symbol (car niins)
                                                   (if (= 0 (car niargs))
                                                      '+ '-))
                                               ,(cons-symbol (car outs)
                                                     (if (= 1 newval) '+ '-)))
                                               (loop ,newstate))) sofar)
				       (cdr niargs) (cdr niins)))
                                 (loop (- state 1) sofar))))
                        `(sm loop ((s ,(sm-expand-init ins)))
                            ,(fix-case `(case s ,@sofar))))))
               (length ins) 1)))))
;
;   A generalized C element is very much like our state machine.
;
(defmac (genc-ckt formals funcup funcdown)
   `(genc-ckt-really ',formals ',funcup ',funcdown))
;
;   We used to use a lambda expression, and this was fairly quick.  Now
;   we use our `language' to interpret it, using macro calls.  This might
;   be slower, but at least it is more consistent.
;
(define (genc-ckt-really formals funcup funcdown)
   (let* ((name (car formals))
          (ins (cadr formals))
          (outs (caddr formals))
          (arity (length ins))
          (p2arity (pow2 arity))
          (fake-macroup (list 'macro 'foo ins funcup 0 0))
          (fake-macrodown (list 'macro 'foo ins funcdown 0 0))
          (p (new-place-unadded))
	  (preup (do-preeval fake-macroup arity p))
	  (predown (do-preeval fake-macrodown arity p)))
      (add-to-environment
         (cons name
            (list 'macro name (append ins outs)
               (list 'complete '()
                  `(when (not (wire-init-known? ,(car outs)))
                     (wire-init-set! ,(car outs)
                        ,(map-to-wire-inits funcup ins)))
		  `(when (and ,(map-to-wire-inits funcup ins)
			      (not (wire-init ,(car outs))))
		     (eval-parse (display-stuff
                                "Pulled up, but output is low in " ',name)))
		  `(when (and ,(map-to-wire-inits funcdown ins)
			      (wire-init ,(car outs)))
		     (eval-parse (display-stuff
                                "Pulled down, but output is high in " ',name)))
                  (let loop ((state (- (* 2 p2arity) 1)) (sofar '()))
                     (if (>= state 0)
                        (let* ((rargs (split-state state (+ arity 1)))
                               (r2args (reverse rargs))
			       (s2 (modulo state p2arity))
                               (up-val (vector-ref preup s2))
                               (down-val (vector-ref predown s2))
                               (init-val (car r2args)))
                           (let loop2 ((i 0) (b 1) (sofar sofar)
				       (niargs rargs) (niins ins))
                              (if (and (< i arity)
                                     (or (= up-val init-val)
                                        (= down-val up-val)))
                                 (let* ((newstate
                                           ((if (= 0 (car niargs)) + -)
                                              state b))
					(na (modulo newstate p2arity))
                                        (newval (vector-ref
						   (if (= init-val 0)
						       preup predown)
						   na)))
                                    (loop2 (+ i 1) (* b 2)
                                       (cons
                                          (if (= newval 0)
                                             `(,state (annotate
                                               ,(cons-symbol (car niins)
                                                   (if (= 0 (car niargs))
                                                      '+ '-)))
                                               (loop ,newstate))
                                             `(,state (annotate
                                               ,(cons-symbol (car niins)
                                                   (if (= 0 (car niargs))
                                                      '+ '-))
                                               ,(cons-symbol (car outs)
                                                     (if (= 0 init-val) '+ '-)))
                                               (loop ,((if (= init-val 0) + -)
						         newstate p2arity))))
					  sofar)
				       (cdr niargs) (cdr niins)))
                                 (loop (- state 1) sofar))))
                        `(sm loop ((s ,(sm-expand-init (append ins outs))))
                            ,(fix-case `(case s ,@sofar))))))
               (length ins) 1)))))
;
;   This function allows us to easily and quickly `construct' circuits
;   without feedback loops out of components with a single output.
;
;   We turn an expression of the form
;
;      (combinational (circ wire-or-circ) dest)
;
;   where wire-or-circ is either a literal name of a wire or else another
;   expression of the above form, into
;
;      (with-wires (temp<1> temp<2> temp<3> ...)
;         (par
;            (innermost-circ inputs output)
;            ...
;            (circ temp1@ temp2@ temp3@)))
;
;   Thus,
;
'      (combinational (fast-and (fast-or a b) c) d)
;
;   becomes
;
;      (with-wires (fast-or<1>)
;         (par
;            (fast-or a b fast-or<1>)
;            (fast-and fast-or<1> c d)))
;
(define (build-eval-circ in-expr)
   (when (not (= (length in-expr) 2))
      (merror "Need two arguments to combinational" in-expr))
   (let ((temps '()) (sofar '()) (n 0))
      ;
      ;   We allow the output to not necessarily be the last
      ;   element of the macro.
      ;
      (define (insert-output circ args o-wire)
         (let* ((mac (verify-ckt? circ *global-assoc*))
		(in (macro-inputs mac)))
            (append (list circ) (my-sublist args 0 in) (list o-wire)
		    (my-sublist args in 1000))))
      (define (next-circ expr)
         (if (or (symbol? expr) (number? expr)) expr
            (let* ((t-wire (if (= (set! n (+ n 1)) 0) (cadr in-expr)
                              (cons-symbol (car expr) '- n '@)))
                   (args (map next-circ (cdr expr))))
               (when (not (eq? t-wire (cadr in-expr)))
	          (set! temps (cons t-wire temps)))
               (set! sofar (cons (insert-output (car expr) args t-wire)
			      sofar))
               t-wire)))
      (next-circ (car in-expr))
      (cons 'with-wires (list temps (cons 'par (reverse sofar))))))
;
(define (pg-and-open)
   (system "pg <foo.grf >foo.ps ; open foo.ps"))
;
;   This next section handles the input of arbitrary safe Petri nets.
;   The only real complexity is generating orbits and making sure the
;   final net is safe.
;
;   To make a place, we simply allocate and return a new place.  We run
;   through the argument list, evaluating each.  If any affect the place
;   other than 
;
(define (make-place args asc)
   (when (null? *global-sub-net*)
      (merror "make-place illegal outside of net" 0))
   (let ((p (new-place-unadded)))
      (set-car! *global-sub-net* (cons p (car *global-sub-net*)))
      (for-each (lambda (exp)
		   (let ((r (build-expr exp p asc)))
		      (when (and (car r) (eqv? (safe-cadr r) 1))
		         (node-marked-set! p #t))))
         args)
      (when (or (node-inedges p) (node-outedges p))
         (merror "Got a non-simple expression in make-place" 0))
      p))
;
(define (make-transition args asc)
   (when (null? *global-sub-net*)
      (merror "make-transition illegal outside of net" 0))
   (let ((t (new-node-unadded 'T)))
      (set-cdr! *global-sub-net* (cons t (cdr *global-sub-net*)))
      (do-annotate '() '() t args asc)
      t))
;
(define (make-edges src args asc ignore-timing?)
   (when args
      (if (eq? (car args) 'ignore-timing)
         (make-edges src (cdr args) asc #t)
	 (begin
            (when (not (list? (car args)))
               (merror "Expected lists as arguments to make-edges" args))
	    (let ((dst (map (lambda (exp)
			       (let ((r (simple-expr exp asc)))
			          (when (or (not (node? r))
					    (and (not (node-type? r 'P))
					         (not (node-type? r 'T))))
                                     (merror
                                 "Expected lists of nodes in make-edges" exp))
                                  r)) (car args))))
               (when (and src dst)
                  (for-each (lambda (src-node)
                               (for-each (lambda (dst-node)
                                            (when (eq? (node-type src-node)
                                                       (node-type dst-node))
                                               (merror
       "Can only connect places to transitions or transitions to places", 0))
                                            (when (not (member dst-node
                                                          (node-out-nodes
                                                             src-node)))
                                               (connect src-node dst-node)
                                               (when ignore-timing?
                                                  (edge-ignore-timing?-set!
                                                     (car
                                                        (node-outedges
							   src-node))
                                                     #t))))
                                  dst))
                     src))
               (make-edges dst (cdr args) asc ignore-timing?))))))
;
(define *sub-net-orbit* 0)
;
(define *sub-net-inc* 0)
;
;   This is the hard part.  We take a particular place and try to locate an
;   orbit based on that place.
;
(define (find-orbit-p p t-stack pyes pcons toks)
   (when (node-marked p)
      (set! toks (+ 1 toks)))
   (if (> toks 1)
      '()
      (find-orbit (append (node-in-nodes p) (node-out-nodes p) t-stack)
         (cons p pyes) pcons toks)))
;
(define (find-orbit t-stack pyes pcon toks)
   (define (order lst)
      (if (< (length lst) 2) lst
         (let loop ((lst lst) (assy '()) (assn '()))
	    (if lst
	       (if (node-orbit (car lst))
	          (loop (cdr lst) (cons (car lst) assy) assn)
	          (loop (cdr lst) assy (cons (car lst) assn)))
	       (append assn assy)))))
   (if (null? t-stack)
      (if (= toks 1) pyes
         (merror "Internal error 2 in find-orbit" 0))
      (let* ((t (car t-stack)) (t-stack (cdr t-stack))
             (pre (node-in-nodes t)) (post (node-out-nodes t))
             (up (length (intersection pre pyes)))
             (down (length (intersection post pyes)))
             (opcon pcon))
         (cond
            ((or (> up 1) (> down 1)) #f)
            ((= up down 1) (find-orbit t-stack pyes pcon toks))
            ((= up down 0) (merror "Internal error in find-orbit" 0))
            ((= up 1)
             (set! pcon (append post pcon))
             (let loop ((trylist (order post)))
                (if trylist
                   (if (member (car trylist) opcon)
                      (loop (cdr trylist))
                      (or (find-orbit-p (car trylist) t-stack pyes pcon toks)
                         (loop (cdr trylist))))
                   #f)))
            ((= down 1)
             (set! pcon (append pre pcon))
             (let loop ((trylist (order pre)))
                (if trylist
                   (if (member (car trylist) opcon)
                      (loop (cdr trylist))
                      (or (find-orbit-p (car trylist) t-stack pyes pcon toks)
                         (loop (cdr trylist))))
                   #f)))))))
;
(define (find-an-orbit-p p)
   (let ((orbit (find-orbit-p p '() '() (list p) 0)))
      (when (null? orbit)
         (merror "net in net statement is not safe" 0))
      (for-each (lambda (p)
                   (when (null? (node-orbit p))
                      (node-orbit-set! p *sub-net-orbit*))) orbit)
      (set! *sub-net-orbit* (+ *sub-net-orbit* *next-orbit*))))
;
(define (make-net place args asc)
   (define (prune-unconnected name lst sofar)
      (if lst
         (if (or (node-inedges (car lst)) (node-outedges (car lst)))
            (cons (car lst) (prune-unconnected name (cdr lst) sofar))
            (prune-unconnected name (cdr lst) (+ sofar 1)))
         (begin
            (when (> sofar 0)
               (display-stuff 'newline "Deleted " sofar " unconnected " name))
            '())))
   (when *global-sub-net*
      (merror "Nested nets not allowed" 0))
   (set! *global-sub-net* (cons '() '()))
   (set! *sub-net-orbit* *current-orbit*)
   (simple-expr (cons 'begin args) asc)
   (when (null? (car *global-sub-net*))
      (merror "No places built in net statement" 0))
   (when (null? (cdr *global-sub-net*))
      (merror "No transitions built in net statement" 0))
;
;   Now we try to assign orbits to all of the places.
;
   (let ((places (prune-unconnected "places" (car *global-sub-net*) 0))
         (transitions (prune-unconnected "transitions"
                         (cdr *global-sub-net*) 0))
         (initial '()))
      (for-each (lambda (p)
                   (node-orbit-set! p '())
                   (when (node-marked p)
                      (set! initial (cons p initial)))) places)
      (when (null? initial)
         (merror "No initially marked places in net statement" 0))
 (display-stuff "Finding orbits for initially marked places" 'newline)
      (for-each (lambda (p)  (display-stuff ".")
                   (find-an-orbit-p p)) initial)
 (display-stuff "Finding orbits for rest of places" 'newline)
      (for-each (lambda (p)
                   (display-stuff ".")
                   (when (not (node-orbit p)) (find-an-orbit-p p))) places)
      (for-each (lambda (p) (add-node-to-graph p)) places)
      (for-each (lambda (t) (add-node-to-graph t)) transitions)
      (for-each (lambda (p) (node-marked-set! p #f)) initial)
      (let ((t (new-trans-from-place place)))
         (for-each (lambda (p) (connect t p)) initial)))
   (set! *global-sub-net* '()))
